# segmentation.py
"""

A DEA customized version of the tiledsegsingle.py module
implemented by the python package RSGISlib.  It has been adapted to run a tiled,
parallel image segmentation across a specified number of cpus.

NOTE: the only function that 99% of users will need to call is "performTiledSegmentation"

Documentation for the RSGISlib Image Segmentation Module can be found here:
https://www.rsgislib.org/rsgislib_segmentation.html


License
-------
The code in this notebook is licensed under the Apache License,
Version 2.0 (https://www.apache.org/licenses/LICENSE-2.0). Digital Earth
Africa data is licensed under the Creative Commons by Attribution 4.0
license (https://creativecommons.org/licenses/by/4.0/).

Contact
-------
If you need assistance, please post a question on the Open Data
Cube Slack channel (http://slack.opendatacube.org/) or on the GIS Stack
Exchange (https://gis.stackexchange.com/questions/ask?tags=open-data-cube)
using the `open-data-cube` tag (you can view previously asked questions
here: https://gis.stackexchange.com/questions/tagged/open-data-cube).

If you would like to report an issue with this script, file one on
Github: https://github.com/GeoscienceAustralia/dea-notebooks/issues/new


.. autosummary::
   :nosignatures:
   :toctree: gen

"""

import glob
import os.path
import os
import rsgislib
from rsgislib.imageutils import tilingutils
from rsgislib.segmentation import segutils
from rsgislib import rastergis
from rsgislib import imageutils
from rsgislib import segmentation
import dill
from pathos.multiprocessing import ProcessingPool as Pool
from osgeo import gdal
from rios import rat
import json
import shutil


class RSGISTiledShepherdSegmentationSingleThread(object):
    """
    A class for running the tiled version of the shepherd segmentation algorithm.
    This can process larger images than the single scene version with a smaller
    memory footprint.

    This version has been adapted to run over multiple cpus. 

    It is not intended that this class will be directly used. Please use the 
    function 'performTiledSegmentation' to call this functionality. 
    """

    def findSegStatsFiles(self, tileImg, segStatsInfo):
        gdalDS = gdal.Open(tileImg, gdal.GA_ReadOnly)
        geotransform = gdalDS.GetGeoTransform()
        if not geotransform is None:
            xMin = geotransform[0]
            yMax = geotransform[3]

            xRes = geotransform[1]
            yRes = geotransform[5]

            width = gdalDS.RasterXSize * xRes
            if yRes < 0:
                yRes = yRes * (-1)
            height = gdalDS.RasterYSize * yRes
            xMax = xMin + width
            yMin = xMax - height

            xCen = xMin + (width / 2)
            yCen = yMin + (height / 2)
        gdalDS = None

        first = True
        minDist = 0.0
        minKCenFile = ""
        minStchStatsFile = ""
        for tileName in segStatsInfo:
            tileXCen = segStatsInfo[tileName]['CENTRE_PT']['X']
            tileYCen = segStatsInfo[tileName]['CENTRE_PT']['Y']
            dist = ((tileXCen - xCen) * (tileXCen - xCen)) + \
                ((tileYCen - yCen) * (tileYCen - yCen))
            if first:
                minKCenFile = segStatsInfo[tileName]['KCENTRES']
                minStchStatsFile = segStatsInfo[tileName]['STRETCHSTATS']
                minDist = dist
                first = False
            elif dist < minDist:
                minKCenFile = segStatsInfo[tileName]['KCENTRES']
                minStchStatsFile = segStatsInfo[tileName]['STRETCHSTATS']
                minDist = dist

        return minKCenFile, minStchStatsFile

    def performStage1Tiling(self, inputImage, tileShp, tilesRat, tilesBase,
                            tilesMetaDIR, tilesImgDIR, tmpDIR, width, height,
                            validDataThreshold):
        tilingutils.createMinDataTiles(inputImage, tileShp, tilesRat, width,
                                       height, validDataThreshold, None, False,
                                       True, tmpDIR)
        tilingutils.createTileMaskImagesFromClumps(tilesRat, tilesBase,
                                                   tilesMetaDIR, "KEA")
        rsgisUtils = rsgislib.RSGISPyUtils()
        dataType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImage)
        tilingutils.createTilesFromMasks(inputImage, tilesBase, tilesMetaDIR,
                                         tilesImgDIR, dataType, 'KEA')

    def performStage1TilesSegmentation(self, tilesImgDIR, stage1TilesSegsDIR,
                                       tmpDIR, tilesBase, tileSegInfoJSON,
                                       strchStatsBase, kCentresBase,
                                       numClustersVal, minPxlsVal, distThresVal,
                                       bandsVal, samplingVal, kmMaxIterVal,
                                       ncpus):
        imgTiles = glob.glob(os.path.join(tilesImgDIR, tilesBase + "*.kea"))

    def performStage1TilesSegmentation(self, tilesImgDIR, stage1TilesSegsDIR,
                                       tmpDIR, tilesBase, tileSegInfoJSON,
                                       strchStatsBase, kCentresBase,
                                       numClustersVal, minPxlsVal, distThresVal,
                                       bandsVal, samplingVal, kmMaxIterVal,
                                       ncpus):
        imgTiles = glob.glob(os.path.join(tilesImgDIR, tilesBase + "*.kea"))

        def threadedTiledImgSeg(imgTile):
            baseName = os.path.splitext(os.path.basename(imgTile))[0]
            tileID = baseName.split('_')[-1]
            clumpsFile = os.path.join(stage1TilesSegsDIR,
                                      baseName + '_segs.kea')
            tmpStatsJSON = os.path.join(tilesImgDIR,
                                        baseName + '_segstats.json')
            strchStatsOutFile = strchStatsBase + "_" + tileID + '.txt'
            kCentresOutFile = kCentresBase + "_" + tileID
            print(clumpsFile)

            try:
                segutils.runShepherdSegmentation(
                    imgTile,
                    clumpsFile,
                    outputMeanImg=None,
                    tmpath=os.path.join(tmpDIR, tileID + '_segstemp'),
                    gdalformat='KEA',
                    noStats=False,
                    noStretch=False,
                    noDelete=False,
                    numClusters=numClustersVal,
                    minPxls=minPxlsVal,
                    distThres=distThresVal,
                    bands=bandsVal,
                    sampling=samplingVal,
                    kmMaxIter=kmMaxIterVal,
                    processInMem=False,
                    saveProcessStats=True,
                    imgStretchStats=strchStatsOutFile,
                    kMeansCentres=kCentresOutFile,
                    imgStatsJSONFile=tmpStatsJSON)
                print(" !!! finished the seg on tile " + imgTile)

            except:
                print(
                    " !&!" + imgTile +
                    " Error: Its likely this tile doesn't have enough of the geotiff inside its extent to run the seg algorithm. I recommend overlaying the tiling shapefile (..._S1Tiles.shp) on the geotiff to check the extents. You can't proceed further until you deal with this as Part 2 will fail."
                )
                pass

        with Pool(ncpus) as p:
            p.map(threadedTiledImgSeg, imgTiles)

        print("completed the stage 1 image seg")

    def defineStage1Boundaries(self, tilesImgDIR, stage1TilesSegBordersDIR,
                               tilesBase):
        segTiles = glob.glob(os.path.join(tilesImgDIR,
                                          tilesBase + "*_segs.kea"))
        for segTile in segTiles:
            baseName = os.path.splitext(os.path.basename(segTile))[0]
            borderMaskFile = os.path.join(stage1TilesSegBordersDIR,
                                          baseName + '_segsborder.kea')
            rastergis.defineBorderClumps(segTile, 'BoundaryClumps')
            rastergis.exportCol2GDALImage(segTile, borderMaskFile, 'KEA',
                                          rsgislib.TYPE_8UINT, 'BoundaryClumps')

    def mergeStage1TilesToOutput(self, inputImage, tilesSegsDIR,
                                 tilesSegsBordersDIR, tilesBase, clumpsImage,
                                 bordersImage):
        segTiles = glob.glob(
            os.path.join(tilesSegsDIR, tilesBase + "*_segs.kea"))
        imageutils.createCopyImage(inputImage, clumpsImage, 1, 0, 'KEA',
                                   rsgislib.TYPE_32UINT)
        segmentation.mergeClumpImages(segTiles, clumpsImage)
        rastergis.populateStats(clumpsImage, True, True)

        tileBorders = glob.glob(
            os.path.join(tilesSegsBordersDIR, tilesBase + "*_segsborder.kea"))
        imageutils.createCopyImage(inputImage, bordersImage, 1, 0, 'KEA',
                                   rsgislib.TYPE_8UINT)
        imageutils.includeImages(bordersImage, tileBorders)
        rastergis.populateStats(bordersImage, True, True)

    def performStage2Tiling(self, inputImage, tileShp, tilesRat, tilesBase,
                            tilesMetaDIR, tilesImgDIR, tmpDIR, width, height,
                            validDataThreshold, bordersImage):
        tilingutils.createMinDataTiles(inputImage, tileShp, tilesRat, width,
                                       height, validDataThreshold, bordersImage,
                                       True, True, tmpDIR)
        tilingutils.createTileMaskImagesFromClumps(tilesRat, tilesBase,
                                                   tilesMetaDIR, "KEA")
        rsgisUtils = rsgislib.RSGISPyUtils()
        dataType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImage)
        tilingutils.createTilesFromMasks(inputImage, tilesBase, tilesMetaDIR,
                                         tilesImgDIR, dataType, 'KEA')

    def performStage2TilesSegmentation(self, tilesImgDIR, tilesMaskedDIR,
                                       tilesSegsDIR, tilesSegBordersDIR, tmpDIR,
                                       tilesBase, s1BordersImage, segStatsInfo,
                                       minPxlsVal, distThresVal, bandsVal,
                                       ncpus):
        rsgisUtils = rsgislib.RSGISPyUtils()
        imgTiles = glob.glob(os.path.join(tilesImgDIR, tilesBase + "*.kea"))
        for imgTile in imgTiles:
            baseName = os.path.splitext(os.path.basename(imgTile))[0]
            maskedFile = os.path.join(tilesMaskedDIR, baseName + '_masked.kea')
            dataType = rsgisUtils.getRSGISLibDataTypeFromImg(imgTile)
            imageutils.maskImage(imgTile, s1BordersImage, maskedFile, 'KEA',
                                 dataType, 0, 0)

        imgTiles = glob.glob(
            os.path.join(tilesMaskedDIR, tilesBase + "*_masked.kea"))

        def stage2threadedTiledImgSeg(imgTile):
            baseName = os.path.splitext(os.path.basename(imgTile))[0]
            clumpsFile = os.path.join(tilesSegsDIR, baseName + '_segs.kea')
            kMeansCentres, imgStretchStats = self.findSegStatsFiles(
                imgTile, segStatsInfo)
            segutils.runShepherdSegmentationPreCalcdStats(
                imgTile,
                clumpsFile,
                kMeansCentres,
                imgStretchStats,
                outputMeanImg=None,
                tmpath=os.path.join(tmpDIR, baseName + '_segstemp'),
                gdalformat='KEA',
                noStats=False,
                noStretch=False,
                noDelete=False,
                minPxls=minPxlsVal,
                distThres=distThresVal,
                bands=bandsVal,
                processInMem=False)

        p = Pool(ncpus)
        p.map(stage2threadedTiledImgSeg, imgTiles)

        segTiles = glob.glob(
            os.path.join(tilesSegsDIR, tilesBase + "*_segs.kea"))
        for segTile in segTiles:
            baseName = os.path.splitext(os.path.basename(segTile))[0]
            borderMaskFile = os.path.join(tilesSegBordersDIR,
                                          baseName + '_segsborder.kea')
            rastergis.defineBorderClumps(segTile, 'BoundaryClumps')
            rastergis.exportCol2GDALImage(segTile, borderMaskFile, 'KEA',
                                          rsgislib.TYPE_8UINT, 'BoundaryClumps')

    def mergeStage2TilesToOutput(self, clumpsImage, tilesSegsDIR,
                                 tilesSegBordersDIR, tilesBase, s2BordersImage):
        segTiles = glob.glob(
            os.path.join(tilesSegsDIR, tilesBase + "*_segs.kea"))
        segmentation.mergeClumpImages(segTiles, clumpsImage)
        rastergis.populateStats(clumpsImage, True, True)

        tileBorders = glob.glob(
            os.path.join(tilesSegBordersDIR, tilesBase + "*_segsborder.kea"))
        imageutils.createCopyImage(clumpsImage, s2BordersImage, 1, 0, 'KEA',
                                   rsgislib.TYPE_8UINT)
        imageutils.includeImages(s2BordersImage, tileBorders)

    def createStage3ImageSubsets(self, inputImage, s2BordersImage,
                                 s3BordersClumps, subsetImgsDIR,
                                 subsetImgsMaskedDIR, subImgBaseName, minSize):
        segmentation.clump(s2BordersImage, s3BordersClumps, 'KEA', True, 0)
        rastergis.populateStats(s3BordersClumps, True, True)

        rastergis.spatialExtent(s3BordersClumps, 'minXX', 'minXY', 'maxXX',
                                'maxXY', 'minYX', 'minYY', 'maxYX', 'maxYY')

        rsgisUtils = rsgislib.RSGISPyUtils()
        dataType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImage)

        ratDS = gdal.Open(s3BordersClumps, gdal.GA_Update)
        minX = rat.readColumn(ratDS, "minXX")
        maxX = rat.readColumn(ratDS, "maxXX")
        minY = rat.readColumn(ratDS, "minYY")
        maxY = rat.readColumn(ratDS, "maxYY")
        Histogram = rat.readColumn(ratDS, "Histogram")
        for i in range(minX.shape[0]):
            if i > 0:
                subImage = os.path.join(subsetImgsDIR,
                                        subImgBaseName + str(i) + '.kea')
                #print( "[" + str(minX[i]) + ", " + str(maxX[i]) + "][" + str(minY[i]) + ", " + str(maxY[i]) + "]" )
                imageutils.subsetbbox(inputImage, subImage, 'KEA', dataType,
                                      minX[i], maxX[i], minY[i], maxY[i])
                if Histogram[i] > minSize:
                    maskedFile = os.path.join(
                        subsetImgsMaskedDIR,
                        subImgBaseName + str(i) + '_masked.kea')
                else:
                    maskedFile = os.path.join(
                        subsetImgsMaskedDIR,
                        subImgBaseName + str(i) + '_burn.kea')
                imageutils.maskImage(subImage, s2BordersImage, maskedFile,
                                     'KEA', dataType, 0, 0)
                rastergis.populateStats(maskedFile, True, False)
        ratDS = None

    def performStage3SubsetsSegmentation(self, subsetImgsMaskedDIR,
                                         subsetSegsDIR, tmpDIR, subImgBaseName,
                                         segStatsInfo, minPxlsVal, distThresVal,
                                         bandsVal, ncpus):
        imgTiles = glob.glob(
            os.path.join(subsetImgsMaskedDIR, subImgBaseName + "*_masked.kea"))

        def stage3threadedTiledImgSeg(imgTile):
            baseName = os.path.splitext(os.path.basename(imgTile))[0]
            clumpsFile = os.path.join(subsetSegsDIR, baseName + '_segs.kea')
            kMeansCentres, imgStretchStats = self.findSegStatsFiles(
                imgTile, segStatsInfo)
            segutils.runShepherdSegmentationPreCalcdStats(
                imgTile,
                clumpsFile,
                kMeansCentres,
                imgStretchStats,
                outputMeanImg=None,
                tmpath=os.path.join(tmpDIR, baseName + '_segstemp'),
                gdalformat='KEA',
                noStats=False,
                noStretch=False,
                noDelete=False,
                minPxls=minPxlsVal,
                distThres=distThresVal,
                bands=bandsVal,
                processInMem=False)

        p = Pool(ncpus)
        p.map(stage3threadedTiledImgSeg, imgTiles)

    def mergeStage3TilesToOutput(self, clumpsImage, subsetSegsDIR,
                                 subsetImgsMaskedDIR, subImgBaseName):
        burnTiles = glob.glob(
            os.path.join(subsetImgsMaskedDIR, subImgBaseName + "*_burn.kea"))
        if len(burnTiles) > 0:
            segmentation.mergeClumpImages(burnTiles, clumpsImage)

        segTiles = glob.glob(
            os.path.join(subsetSegsDIR, subImgBaseName + "*_segs.kea"))
        segmentation.mergeClumpImages(segTiles, clumpsImage)
        rastergis.populateStats(clumpsImage, True, True)


def performTiledSegmentation(inputImage,
                             clumpsImage,
                             tmpDIR='segtmp',
                             tileWidth=2000,
                             tileHeight=2000,
                             validDataThreshold=0.3,
                             numClusters=60,
                             minPxls=100,
                             distThres=100,
                             bands=None,
                             sampling=100,
                             kmMaxIter=200,
                             ncpus=1):
    """
    Utility function to call the segmentation algorithm of Shepherd et al. (2014)
    using the tiled process outlined in Clewley et al (2015). Adapted here to run tiles
    across multiple cpus. Use this function to conduct image segmentation on very large
    geotiffs.

    Parameters
    ----------
    inputImage : str
        is a string containing the name of the input file.
    clumpsImage : str
        is a string containing the name of the output clump file.
    tmpath : str 
        is a file path for intermediate files (default is to create a directory 'segtmp').
        If path does current not exist then it will be created and deleted afterwards.
    tileWidth : int
        is an int specifying the width of the tiles used for processing (Default 2000)
    tileHeight : int
        is an int specifying the height of the tiles used for processing (Default 2000)
    validDataThreshold : float
        is a float (value between 0 - 1) used to specify the amount of valid image
        pixels (i.e., not a no data value of zero) are within a tile. Tiles failing to meet this
        threshold are merged with ones which do (Default 0.3).
    numClusters : int 
        is an int which specifies the number of clusters within the KMeans clustering (default = 60).
    minPxls : int
        is an int which specifies the minimum number pixels within a segments (default = 100).
    distThres : int 
        specifies the distance threshold for joining the segments (default = 100, set to large number to turn off this option).
    bands : array-like
        is an array providing a subset of image bands to use (default is None to use all bands).
    sampling : 
        specify the subsampling of the image for the data used within the KMeans (default = 100; 1 == no subsampling).
    kmMaxIter : 
        maximum iterations for KMeans (Default 200).

    Returns
    -------
    Segmented .kea file stored on disk at the location of 'clumpsImage'
    
    """
    createdTmp = False
    if not os.path.exists(tmpDIR):
        os.makedirs(tmpDIR)
        createdTmp = True
    rsgisUtils = rsgislib.RSGISPyUtils()
    uidStr = rsgisUtils.uidGenerator()

    baseName = os.path.splitext(os.path.basename(inputImage))[0] + "_" + uidStr

    tileSegInfo = os.path.join(tmpDIR, baseName + '_seginfo.json')
    segStatsDIR = os.path.join(tmpDIR, 'segStats_' + uidStr)
    strchStatsBase = os.path.join(segStatsDIR, baseName + '_stch')
    kCentresBase = os.path.join(segStatsDIR, baseName + '_kcentres')
    if not os.path.exists(segStatsDIR):
        os.makedirs(segStatsDIR)

    tiledSegObj = RSGISTiledShepherdSegmentationSingleThread()

    ######################## STAGE 1 #######################
    # Stage 1 Parameters (Internal)
    stage1TileShp = os.path.join(tmpDIR, baseName + '_S1Tiles.shp')
    stage1TileRAT = os.path.join(tmpDIR, baseName + '_S1Tiles.kea')
    stage1TilesBase = baseName + '_S1Tile'
    stage1TilesImgDIR = os.path.join(tmpDIR, 's1tilesimgs_' + uidStr)
    stage1TilesMetaDIR = os.path.join(tmpDIR, 's1tilesmeta_' + uidStr)
    stage1TilesSegsDIR = os.path.join(tmpDIR, 's1tilessegs_' + uidStr)
    stage1TilesSegBordersDIR = os.path.join(tmpDIR,
                                            's1tilessegborders_' + uidStr)
    stage1BordersImage = os.path.join(tmpDIR, baseName + '_S1Borders.kea')

    if not os.path.exists(stage1TilesImgDIR):
        os.makedirs(stage1TilesImgDIR)
    if not os.path.exists(stage1TilesSegsDIR):
        os.makedirs(stage1TilesSegsDIR)
    if not os.path.exists(stage1TilesSegBordersDIR):
        os.makedirs(stage1TilesSegBordersDIR)
    if not os.path.exists(stage1TilesMetaDIR):
        os.makedirs(stage1TilesMetaDIR)

    # Initial Tiling
    tiledSegObj.performStage1Tiling(inputImage, stage1TileShp, stage1TileRAT,
                                    stage1TilesBase, stage1TilesMetaDIR,
                                    stage1TilesImgDIR,
                                    os.path.join(tmpDIR, 's1tilingtemp'),
                                    tileWidth, tileHeight, validDataThreshold)

    # Perform Segmentation
    print('Performing Stage 1 seg...multithreaded with ' + str(ncpus) + " cpus")
    tiledSegObj.performStage1TilesSegmentation(
        stage1TilesImgDIR, stage1TilesSegsDIR, tmpDIR, stage1TilesBase,
        tileSegInfo, strchStatsBase, kCentresBase, numClusters, minPxls,
        distThres, bands, sampling, kmMaxIter, ncpus)

    # grab the stats for each tile and add to json
    tileStatsFiles = dict()
    for file in os.listdir(stage1TilesImgDIR):
        if file.endswith(".json"):
            baseName = os.path.splitext(os.path.basename(file))[0][:-9]
            with open(stage1TilesImgDIR + "/" + file, 'r') as f:
                jsonStrData = f.read()
            segStatsInfo = json.loads(jsonStrData)
            tileStatsFiles[baseName] = segStatsInfo

    with open(tileSegInfo, 'w') as outfile:
        json.dump(tileStatsFiles,
                  outfile,
                  sort_keys=True,
                  indent=4,
                  separators=(',', ': '),
                  ensure_ascii=False)

    # Define Boundaries
    print('defining stage 1 boundaries')
    tiledSegObj.defineStage1Boundaries(stage1TilesSegsDIR,
                                       stage1TilesSegBordersDIR,
                                       stage1TilesBase)

    print('Merging stage 1 segmented tiles')
    # Merge the Initial Tiles
    tiledSegObj.mergeStage1TilesToOutput(inputImage, stage1TilesSegsDIR,
                                         stage1TilesSegBordersDIR,
                                         stage1TilesBase, clumpsImage,
                                         stage1BordersImage)

    shutil.rmtree(stage1TilesImgDIR)
    shutil.rmtree(stage1TilesSegsDIR)
    shutil.rmtree(stage1TilesSegBordersDIR)
    shutil.rmtree(stage1TilesMetaDIR)
    ########################################################

    with open(tileSegInfo, 'r') as f:
        jsonStrData = f.read()
    segStatsInfo = json.loads(jsonStrData)

    ######################## STAGE 2 #######################
    # Stage 2 Parameters (Internal)
    stage2TileShp = os.path.join(tmpDIR, baseName + '_S2Tiles.shp')
    stage2TileRAT = os.path.join(tmpDIR, baseName + '_S2Tiles.kea')
    stage2TilesBase = baseName + '_S2Tile'
    stage2TilesImgDIR = os.path.join(tmpDIR, 's2tilesimg_' + uidStr)
    stage2TilesMetaDIR = os.path.join(tmpDIR, 's2tilesmeta_' + uidStr)
    stage2TilesImgMaskedDIR = os.path.join(tmpDIR, 's2tilesimgmask_' + uidStr)
    stage2TilesSegsDIR = os.path.join(tmpDIR, 's2tilessegs_' + uidStr)
    stage2TilesSegBordersDIR = os.path.join(tmpDIR,
                                            's2tilessegborders_' + uidStr)
    stage2BordersImage = os.path.join(tmpDIR, baseName + '_S2Borders.kea')

    if not os.path.exists(stage2TilesImgDIR):
        os.makedirs(stage2TilesImgDIR)
    if not os.path.exists(stage2TilesMetaDIR):
        os.makedirs(stage2TilesMetaDIR)
    if not os.path.exists(stage2TilesImgMaskedDIR):
        os.makedirs(stage2TilesImgMaskedDIR)
    if not os.path.exists(stage2TilesSegsDIR):
        os.makedirs(stage2TilesSegsDIR)
    if not os.path.exists(stage2TilesSegBordersDIR):
        os.makedirs(stage2TilesSegBordersDIR)

    # Perform offset tiling
    print('Performing stage2 offset tilings')
    tiledSegObj.performStage2Tiling(inputImage, stage2TileShp, stage2TileRAT,
                                    stage2TilesBase, stage2TilesMetaDIR,
                                    stage2TilesImgDIR,
                                    os.path.join(tmpDIR, 's2tilingtemp'),
                                    tileWidth, tileHeight, validDataThreshold,
                                    stage1BordersImage)

    print('Performing segmentation of stage2 offset tiles')
    # Perform Segmentation of the Offset Tiles
    tiledSegObj.performStage2TilesSegmentation(
        stage2TilesImgDIR, stage2TilesImgMaskedDIR, stage2TilesSegsDIR,
        stage2TilesSegBordersDIR, tmpDIR, stage2TilesBase, stage1BordersImage,
        segStatsInfo, minPxls, distThres, bands, ncpus)

    # Merge in the next set of boundaries
    print('merging next set of boundaries')
    tiledSegObj.mergeStage2TilesToOutput(clumpsImage, stage2TilesSegsDIR,
                                         stage2TilesSegBordersDIR,
                                         stage2TilesBase, stage2BordersImage)

    shutil.rmtree(stage2TilesImgDIR)
    shutil.rmtree(stage2TilesMetaDIR)
    shutil.rmtree(stage2TilesImgMaskedDIR)
    shutil.rmtree(stage2TilesSegsDIR)
    shutil.rmtree(stage2TilesSegBordersDIR)
    ########################################################

    ######################## STAGE 3 #######################
    print('starting stage 3')
    # Stage 3 Parameters (Internal)
    stage3BordersClumps = os.path.join(tmpDIR,
                                       baseName + '_S3BordersClumps.kea')
    stage3SubsetsDIR = os.path.join(tmpDIR, 's3subsetimgs_' + uidStr)
    stage3SubsetsMaskedDIR = os.path.join(tmpDIR, 's3subsetimgsmask_' + uidStr)
    stage3SubsetsSegsDIR = os.path.join(tmpDIR, 's3subsetsegs_' + uidStr)
    stage3Base = baseName + '_S3Subset'

    if not os.path.exists(stage3SubsetsDIR):
        os.makedirs(stage3SubsetsDIR)
    if not os.path.exists(stage3SubsetsMaskedDIR):
        os.makedirs(stage3SubsetsMaskedDIR)
    if not os.path.exists(stage3SubsetsSegsDIR):
        os.makedirs(stage3SubsetsSegsDIR)

    # Create the final boundary image subsets
    print('creating final image subsets')
    tiledSegObj.createStage3ImageSubsets(inputImage, stage2BordersImage,
                                         stage3BordersClumps, stage3SubsetsDIR,
                                         stage3SubsetsMaskedDIR, stage3Base,
                                         minPxls)

    # Perform Segmentation of the stage 3 regions
    print("Performing Segmentation of the stage 3 regions")
    tiledSegObj.performStage3SubsetsSegmentation(stage3SubsetsMaskedDIR,
                                                 stage3SubsetsSegsDIR, tmpDIR,
                                                 stage3Base, segStatsInfo,
                                                 minPxls, distThres, bands,
                                                 ncpus)

    # Merge the stage 3 regions into the final clumps image
    print('merging stage 3 regions into the final clumps image')
    tiledSegObj.mergeStage3TilesToOutput(clumpsImage, stage3SubsetsSegsDIR,
                                         stage3SubsetsMaskedDIR, stage3Base)

    shutil.rmtree(stage3SubsetsDIR)
    shutil.rmtree(stage3SubsetsMaskedDIR)
    shutil.rmtree(stage3SubsetsSegsDIR)
    ########################################################

    shutil.rmtree(segStatsDIR)
    rsgisUtils = rsgislib.RSGISPyUtils()
    rsgisUtils.deleteFileWithBasename(stage1BordersImage)
    rsgisUtils.deleteFileWithBasename(stage2BordersImage)
    rsgisUtils.deleteFileWithBasename(stage3BordersClumps)
    rsgisUtils.deleteFileWithBasename(stage1TileShp)
    rsgisUtils.deleteFileWithBasename(stage1TileRAT)
    rsgisUtils.deleteFileWithBasename(stage2TileShp)
    rsgisUtils.deleteFileWithBasename(stage2TileRAT)
    os.remove(tileSegInfo)
    if createdTmp:
        shutil.rmtree(tmpDIR)
    print('Done!')



'''
Description: Functions for easily defining widgets for Digital Earth 
Australia notebooks. These functions are directly modified from 
originals developed by Digital Earth Africa: 
https://github.com/digitalearthafrica/deafrica-sandbox-notebooks/blob/main/Tools/deafrica_tools/app/widgetconstructors.py

License: The code in this notebook is licensed under the Apache License, 
Version 2.0 (https://www.apache.org/licenses/LICENSE-2.0). Digital Earth 
Australia data is licensed under the Creative Commons by Attribution 4.0 
license (https://creativecommons.org/licenses/by/4.0/).

Contact: If you need assistance, post a question on the Open Data Cube 
Slack channel (http://slack.opendatacube.org/) or the GIS Stack Exchange 
(https://gis.stackexchange.com/questions/ask?tags=open-data-cube) using 
the `open-data-cube` tag (you can view previously asked questions here: 
https://gis.stackexchange.com/questions/tagged/open-data-cube). 

If you would like to report an issue with this script, you can file one 
on Github (https://github.com/GeoscienceAustralia/dea-notebooks/issues/new).

Last modified: June 2022
'''

import ipyleaflet as leaflet
from ipyleaflet import LayersControl
import ipywidgets as widgets
from traitlets import Unicode


def create_datepicker(description='', value=None, layout={'width': '85%'}):
    '''
    Create a DatePicker widget
    
    Last modified: October 2021
    
    Parameters
    ----------
    description : string
        descirption label to attach
    layout : dictionary
        any layout commands for the widget
        
    Returns
    -------
    date_picker : ipywidgets.widgets.widget_date.DatePicker
        
    '''

    date_picker = widgets.DatePicker(description=description,
                                     layout=layout,
                                     disabled=False,
                                     value=value)

    return date_picker


def create_inputtext(value,
                     placeholder,
                     description="",
                     layout={'width': '85%'}):
    '''
    Create a Text widget
    
    Last modified: October 2021
    
    Parameters
    ----------
    value : string
        initial value of the widget
    placeholder : string
        placeholder text to display to the user before intput
    description : string
        descirption label to attach
    layout : dictionary
        any layout commands for the widget
        
    Returns
    -------
    input_text : ipywidgets.widgets.widget_string.Text
        
    '''

    input_text = widgets.Text(value=value,
                              placeholder=placeholder,
                              description=description,
                              layout=layout,
                              disabled=False)

    return input_text


def create_boundedfloattext(value,
                            min_val,
                            max_val,
                            step_val,
                            description="",
                            layout={'width': '85%'}):
    '''
    Create a BoundedFloatText widget
    
    Last modified: October 2021
    
    Parameters
    ----------
    value : float
        initial value of the widget
    min_val : float
        minimum allowed value for the float
    max_val : float
        maximum allowed value for the float
    step_val : float
        allowed increment for the float
    description : string
        descirption label to attach
    layout : dictionary
        any layout commands for the widget
        
    Returns
    -------
    float_text : ipywidgets.widgets.widget_float.BoundedFloatText
        
    '''

    float_text = widgets.BoundedFloatText(
        value=value,
        min=min_val,
        max=max_val,
        step=step_val,
        description=description,
        layout=layout,
        disabled=False,
    )

    return float_text


def create_dropdown(options, value, description="", layout={'width': '85%'}):
    '''
    Create a Dropdown widget
    
    Last modified: October 2021
    
    Parameters
    ----------
    options : list
        a list of options for the user to select from
    value : string
        initial value of the widget
    description : string
        descirption label to attach
    layout : dictionary
        any layout commands for the widget
        
    Returns
    -------
    dropdown : ipywidgets.widgets.widget_selection.Dropdown
        
    '''

    dropdown = widgets.Dropdown(
        options=options,
        value=value,
        description=description,
        layout=layout,
        disabled=False,
    )

    return dropdown


def create_checkbox(value, description="", layout={'width': '85%'}):
    '''
    Create a Checkbox widget
    
    Last modified: January 2022
    
    Parameters
    ----------
    value : string
        initial value of the widget; True or False
    description : string
        description label to attach
    layout : dictionary
        any layout commands for the widget
        
    Returns
    -------
    dropdown : ipywidgets.widgets.widget_selection.Dropdown
        
    '''

    checklist = widgets.Checkbox(value=value,
                                 description=description,
                                 layout=layout,
                                 disabled=False,
                                 indent=False)

    return checklist


def create_html(value):
    '''
    Create a HTML widget
    
    Last modified: October 2021
    
    Parameters
    ----------
    value : string
        HTML text to display
        
    Returns
    -------
    html : ipywidgets.widgets.widget_string.HTML
        
    '''

    html = widgets.HTML(value=value,)

    return html


def create_map(map_center=(-28, 135),
               zoom_level=4,
               basemap=leaflet.basemaps.OpenStreetMap.Mapnik,
               basemap_name='Open Street Map',
               **kwargs):
    '''
    Create an interactive ipyleaflet map
    
    Last modified: October 2021
    
    Parameters
    ----------
    map_center : tuple
        A tuple containing the latitude and longitude to focus on.
        Defaults to center of Australia (-28, 135)
    zoom_level : integer
        Zoom level for the map
        Defaults to 4 to view all of Australia
    basemap : ipyleaflet basemap (dict)
        Basemap to use, can be any from https://ipyleaflet.readthedocs.io/en/latest/api_reference/basemaps.html
        Defaults to Open Street Map (basemaps.OpenStreetMap.Mapnik)
    basemap_name : string
        Layer name for the basemap
        
    Returns
    -------
    m : ipyleaflet.leaflet.Map
        interactive ipyleaflet map
        
    '''

    basemap_tiles = leaflet.basemap_to_tiles(basemap)
    basemap_tiles.name = basemap_name

    m = leaflet.Map(center=map_center,
                    zoom=zoom_level,
                    basemap=basemap_tiles,
                    scroll_wheel_zoom=True,
                    **kwargs)

    return m


def create_dea_wms_layer(product, date, **params):
    '''
    Create a Digital Earth Australia WMS layer to add to a map
    
    Last modified: October 2021
    
    Parameters
    ----------
    product : string
        The Digital Earth Australia product to load
        (e.g. 'ga_ls8c_nbart_gm_cyear_3')
    date : string (yyyy-mm-dd format)
        The date to load the product for
        
    Returns
    -------
    time_wms : ipyleaflet WMS layer
        
    '''

    # Load DEA WMS
    class TimeWMSLayer(leaflet.WMSLayer):
        time = Unicode("").tag(sync=True, o=True)

    time_wms = TimeWMSLayer(
        url="https://ows.dea.ga.gov.au/",
        layers=product,
        time=date,
        format="image/png",
        transparent=True,
        attribution="Digital Earth Australia",
        **params
    )

    return time_wms


def create_drawcontrol(
    draw_controls=[
        'rectangle', 'polygon', 'circle', 'polyline', 'marker', 'circlemarker'
    ],
    rectangle_options={},
    polygon_options={},
    circle_options={},
    polyline_options={},
    marker_options={},
    circlemarker_options={},
):
    '''
    Create a draw control widget to add to ipyleaflet maps
    
    Last modified: October 2021
    
    Parameters
    ----------
    draw_controls : list
        List of draw controls to add to the map. Defaults to adding all
        Viable options are 'rectangle', 'polygon', 'circle', 'polyline', 
        'marker', 'circlemarker'.
    rectangle_options : dict
        Options to customise the appearence of the relevant shape. 
        Leave blank for default styling.
    polygon_options : dict
        Options to customise the appearence of the relevant shape. 
        Leave blank for default styling.
    circle_options : dict
        Options to customise the appearence of the relevant shape. 
        Leave blank for default styling.
    polyline_options : dict
        Options to customise the appearence of the relevant shape. 
        Leave blank for default styling.
    marker_options : dict
        Options to customise the appearence of the relevant shape. 
        Leave blank for default styling.
    circlemarker_options : dict
        Options to customise the appearence of the relevant shape. 
        Leave blank for default styling.
    
        
    Returns
    -------
    draw_control : ipyleaflet.leaflet.DrawControl
        
    '''

    # Set defualt DE Africa styling options for polygons
    default_shapeoptions = {
        "color": "#FFFFFF",
        "opacity": 0.8,
        "fillColor": "#336699",
        "fillOpacity": 0.4,
    }
    default_drawerror = {
        "color": "#FF6633",
        "message": "Drawing error, clear all and try again"
    }

    # Set draw control appearence to DE Africa defaults
    # Do this if user has requested a control, but has not provided a corresponding options dict

    if ('rectangle' in draw_controls) and (not rectangle_options):
        rectangle_options = {"shapeOptions": default_shapeoptions}

    if ('polygon' in draw_controls) and (not polygon_options):
        polygon_options = {
            "shapeOptions": default_shapeoptions,
            "drawError": default_drawerror,
            "allowIntersection": False,
        }

    if ('circle' in draw_controls) and (not circle_options):
        circle_options = {"shapeOptions": default_shapeoptions}

    if ('polyline' in draw_controls) and (not polyline_options):
        polyline_options = {"shapeOptions": default_shapeoptions}

    if ('marker' in draw_controls) and (not marker_options):
        marker_options = {'shapeOptions': {'opacity': 1.0}}

    if ('circlemarker' in draw_controls) and (not circlemarker_options):
        circlemarker_options = {"shapeOptions": default_shapeoptions}

    # Instantiate draw control and add options
    draw_control = leaflet.DrawControl()
    draw_control.rectangle = rectangle_options
    draw_control.polygon = polygon_options
    draw_control.marker = marker_options
    draw_control.circle = circle_options
    draw_control.circlemarker = circlemarker_options
    draw_control.polyline = polyline_options

    return draw_control



# -*- coding: utf-8 -*-
"""
Image export widget, which can be used to interactively select and 
export satellite imagery from multiple DEA products.
"""

# Import required packages
import fiona
import sys
import datacube
import warnings
import matplotlib.pyplot as plt
from datacube.utils.geometry import CRS
from ipyleaflet import (
    WMSLayer,
    basemaps,
    basemap_to_tiles,
    Map,
    DrawControl,
    WidgetControl,
    LayerGroup,
    LayersControl,
    GeoData,
)
from traitlets import Unicode
from ipywidgets import (
    GridspecLayout,
    Button,
    Layout,
    HBox,
    VBox,
    HTML,
    Output,
)
import json
import itertools
import numpy as np
import geopandas as gpd
from io import BytesIO
import ipywidgets as widgets
import datetime
from skimage import exposure
from skimage.filters import unsharp_mask

from datacube.utils import masking
from datacube.utils.geometry import Geometry
import dea_tools.app.widgetconstructors as deawidgets
from dea_tools.dask import create_local_dask_cluster
from dea_tools.spatial import reverse_geocode
from dea_tools.datahandling import pan_sharpen_brovey


# WMS params and satellite style bands
sat_params = {
    'ga_ls_ard_3': {
        'products': ['ga_ls5t_ard_3', 'ga_ls7e_ard_3', 'ga_ls8c_ard_3'],
        'styles': {
            'True colour':
                ('true_colour', ['nbart_red', 'nbart_green', 'nbart_blue']),
            'False colour':
                ('false_colour', ['nbart_swir_1', 'nbart_nir', 'nbart_green'])
        }
    },
    's2_ard_granule_nbar_t': {
        'products': ['s2a_ard_granule', 's2b_ard_granule'],
        'styles': {
            'True colour':
                ('simple_rgb', ['nbart_red', 'nbart_green', 'nbart_blue']),
            'False colour': ('infrared_green',
                             ['nbart_swir_2', 'nbart_nir_1', 'nbart_green'])
        }
    },
    's2_nrt_granule_nbar_t': {
        'products': ['s2a_nrt_granule', 's2b_nrt_granule'],
        'styles': {
            'True colour':
                ('simple_rgb', ['nbart_red', 'nbart_green', 'nbart_blue']),
            'False colour': ('infrared_green',
                             ['nbart_swir_2', 'nbart_nir_1', 'nbart_green'])
        }
    }
}


def make_box_layout():
    return Layout(
        #         border='solid 1px black',
        margin='0px 10px 10px 0px',
        padding='5px 5px 5px 5px',
        width='100%',
        height='100%',
    )


def create_expanded_button(description, button_style):
    return Button(
        description=description,
        button_style=button_style,
        layout=Layout(width="auto", height="auto"),
    )


def update_map_layers(self):
    """
    Updates map to add new DEA layers, styles or basemap when selected
    using menu options. Triggers data reload by resetting load params 
    and output arrays.
    """

    # Clear data load params to trigger data re-load
    self.rgb_array = None
    self.sensor = None
    self.load_params = None
    self.query_params = None

    # Clear all layers and add basemap
    self.map_layers.clear_layers()
    self.map_layers.add_layer(self.basemap)

    # Get style name for specific satellite sensor
    style = sat_params[self.dealayer]['styles'][self.style][0]

    # Add DEA layers over the top of the basemap
    dea_layer = deawidgets.create_dea_wms_layer(self.dealayer,
                                                self.date,
                                                styles=style)
    self.map_layers.add_layer(dea_layer)
    
    
    
def extract_data(self):
    
    # Connect to datacube database
    dc = datacube.Datacube(app="Exporting satellite images")

    # Configure local dask cluster
    client = create_local_dask_cluster(return_client=True,
                                       display_client=True)

    # Convert to geopolygon
    geopolygon = Geometry(geom=self.gdf_drawn.geometry[0],
                          crs=self.gdf_drawn.crs)

    # Create query after adjusting interval time to UTC by
    # adding a UTC offset of -10 hours. This results issues
    # on the east coast of Australia where satelite overpasses
    # can occur on either side of 24:00 hours UTC
    start_date = np.datetime64(self.date) - np.timedelta64(10, 'h')
    end_date = np.datetime64(self.date) + np.timedelta64(14, 'h')
    self.query_params = {
        'time': (str(start_date), str(end_date)),
        'geopolygon': geopolygon
    }

    # Find matching datasets
    dss = [
        dc.find_datasets(product=i, **self.query_params)
        for i in sat_params[self.dealayer]['products']
    ]
    dss = list(itertools.chain.from_iterable(dss))
    
    # If data is found
    if len(dss) > 0:

        # Get CRS
        crs = str(dss[0].crs)

        # Get sensor (try/except to account for different S2 NRT metadata)
        try:
            sensor = dss[0].metadata_doc['properties'][
                'eo:platform'].capitalize()
        except:
            sensor = dss[0].metadata_doc['platform']['code'].capitalize(
            )
        self.sensor = sensor[0:-1].replace(
            '_', '-') + sensor[-1].capitalize()

        # Meets pansharpening requirements
        can_pansharpen = self.style == 'True colour' and self.sensor in [
            'Landsat-7', 'Landsat-8'
        ]

        # Set up load params
        if self.pansharpen and can_pansharpen:
            self.load_params = {
                'measurements':
                    sat_params[self.dealayer]['styles'][self.style][1] +
                    ['nbart_panchromatic'],
                'resolution': (-self.resolution, self.resolution),
                'align': (7.5, 7.5),
                'output_crs':
                    crs
            }

        else:

            # Use resolution if provided, otherwise use default
            if self.resolution:
                sat_params[self.dealayer]['resolution'] = (
                    -self.resolution, self.resolution)

            self.load_params = {
                'measurements':
                    sat_params[self.dealayer]['styles'][self.style][1],
                'resolution': (-self.resolution, self.resolution),
                'output_crs':
                    crs
            }

        # Load data from datasets
        print(f'Loading {self.sensor} satellite data')
        ds = dc.load(datasets=dss,
                     resampling='bilinear',
                     group_by='solar_day',
                     dask_chunks={
                         'time': 1,
                         'x': 2048,
                         'y': 2048
                     },
                     **self.load_params,
                     **self.query_params)
        ds = masking.mask_invalid_data(ds)

        # Create plain numpy array, optionally after pansharpening
        if self.pansharpen and can_pansharpen:

            # Perform Brovey pan-sharpening and return three numpy.arrays
            print(
                f'Pansharpening {self.sensor} image to 15 m resolution')
            red_sharpen, green_sharpen, blue_sharpen = pan_sharpen_brovey(
                band_1=ds.nbart_red,
                band_2=ds.nbart_green,
                band_3=ds.nbart_blue,
                pan_band=ds.nbart_panchromatic)
            rgb_array = np.vstack(
                [red_sharpen, green_sharpen, blue_sharpen])

            # Match histogram to original array
            from skimage.exposure import match_histograms
            original_array = ds.drop(
                'nbart_panchromatic').to_array().squeeze('time').values
            rgb_array = match_histograms(image=rgb_array,
                                         reference=original_array,
                                         multichannel=True)

        # If pansharpening is requested but not possible, deactivate 
        # pansharpening and reset to 30 m resolution
        elif self.pansharpen and not can_pansharpen:
            print(
                '\nUnable to pansharpen; reverting to 30 m resolution'
            )
            self.checkbox_pansharpen.value = False
            self.text_resolution.disabled = False
            self.text_resolution.value = 30
            rgb_array = ds.isel(time=0).to_array().values

        else:
            rgb_array = ds.isel(time=0).to_array().values

        # Transpose numpy array
        rgb_array = np.transpose(rgb_array, axes=[1, 2, 0])
    
    # Else if no data is returned, return None
    else:
        rgb_array = None
        
    # Close down the dask client
    client.close()
        
    return rgb_array


def plot_data(self, fname):
    
    # Data to plot
    to_plot = self.rgb_array

    # If percentile stretch is supplied, calculate vmin and vmax
    # from percentiles
    if self.percentile_stretch:
        vmin, vmax = np.nanpercentile(to_plot, self.percentile_stretch)
    else:
        vmin, vmax = self.vmin, self.vmax

    # Raise by power to dampen bright features and enhance dark.
    # Raise vmin and vmax by same amount to ensure proper stretch
    if self.power < 1.0:
        with self.status_info:
            print(f'\nApplying power transformation ({self.power})')
        to_plot = to_plot ** self.power
        vmin, vmax = vmin ** self.power, vmax ** self.power
        
    # Rescale/stretch imagery between vmin and vmax
    to_plot = exposure.rescale_intensity(to_plot.astype(float),
                                         in_range=(vmin, vmax),
                                         out_range=(0.0, 1.0))
    
    # Unsharp mask
    if self.unsharp_mask:
        with self.status_info:
            print(f'\nApplying unsharp masking with {self.unsharp_mask_radius} '
                  f'radius and {self.unsharp_mask_amount} amount')
        to_plot = unsharp_mask(to_plot, 
                               radius=self.unsharp_mask_radius, 
                               amount=self.unsharp_mask_amount)       

    
    # Create figure with aspect ratio of data
    fig = plt.figure(dpi=100)
    fig.set_size_inches(10, 10 / (to_plot.shape[1] / to_plot.shape[0]))

    # Remove axes to plot just array data
    ax = plt.Axes(fig, [0., 0., 1., 1.],)
    ax.set_axis_off()
    fig.add_axes(ax)

    # Add data to plot
    ax.imshow(to_plot)
        
    # If a min DPI is specified and image is less than DPI
    if (self.dpi > 0) and (to_plot.shape[1] < self.dpi * 10):
        
        # Export figure to file using exact DPI
        with self.status_info:
            print(f'\nExporting image at {self.dpi} DPI')
        fig.savefig(fname.replace('resolution', 
                                  f'resolution, {self.dpi} DPI'), 
                    dpi=self.dpi)
        
    # If no minumum DPI is specified, export raw array data in native
    # resolution
    else:
        plt.imsave(fname=fname,
                   arr=np.ascontiguousarray(to_plot),
                   format=self.output_format)
        
    # Add plot preview below map and finish
    plt.show()
    with self.status_info:
        print(f'\nImage successfully exported to:\n{fname}.')
    



class imageexport_app(HBox):

    def __init__(self):
        super().__init__()

        ######################
        # INITIAL ATTRIBUTES #
        ######################

        # Basemap
        self.basemap_list = [
            ("Open Street Map",
             basemap_to_tiles(basemaps.OpenStreetMap.Mapnik)),
            ("ESRI World Imagery",
             basemap_to_tiles(basemaps.Esri.WorldImagery)),
        ]
        self.basemap = self.basemap_list[0][1]

        # Satellite data
        self.date = "2021-12-31"
        self.dealayer_list = [
            ("Landsat", 'ga_ls_ard_3'),
            ("Sentinel-2", 's2_ard_granule_nbar_t'),
            ("Sentinel-2 NRT", 's2_nrt_granule_nbar_t'),
        ]
        self.dealayer = self.dealayer_list[0][1]

        # Styles
        self.styles_list = ["True colour", "False colour"]
        self.style = self.styles_list[0]

        # Analysis params
        self.resolution = 30
        self.pansharpen = False
        self.standardise_name = False
        self.vmin = 50
        self.vmax = 3000
        self.percentile_stretch = None  # (1, 99)
        self.power = 1.0
        self.output_list = [("JPG", 'jpg'), ("PNG", 'png')]
        self.output_format = self.output_list[0][1]        
        self.unsharp_mask = False
        self.unsharp_mask_radius = 20
        self.unsharp_mask_amount = 0.3
        self.max_size = False
        self.dpi = 0

        # Drawing params
        self.target = None
        self.action = None
        self.gdf_drawn = None

        # Data load params
        self.rgb_array = None
        self.sensor = None
        self.load_params = None
        self.query_params = None

        ##################
        # HEADER FOR APP #
        ##################

        # Create the Header widget
        header_title_text = "<h3>Digital Earth Australia satellite image export</h3>"
        instruction_text = ("<p>Select the desired satellite data, imagery "
                            "date and image style, zoom in until satellite "
                            "imagery appears on the map, then draw a "
                            "rectangle to select an area of imagery to "
                            "export as a high-resolution image file.</p>")
        self.header = deawidgets.create_html(
            f"{header_title_text}{instruction_text}")
        self.header.layout = make_box_layout()

        #####################################
        # HANDLER FUNCTION FOR DRAW CONTROL #
        #####################################

        # Define the action to take once something is drawn on the map
        def update_geojson(target, action, geo_json):

            # Get data from action
            self.action = action

            # Clear data load params to trigger data re-load
            self.rgb_array = None
            self.sensor = None
            self.load_params = None
            self.query_params = None

            # Convert data to geopandas
            json_data = json.dumps(geo_json)
            binary_data = json_data.encode()
            io = BytesIO(binary_data)
            io.seek(0)
            gdf = gpd.read_file(io)
            gdf.crs = "EPSG:4326"

            # Convert to Albers and compute area
            gdf_drawn_albers = gdf.copy().to_crs("EPSG:3577")
            m2_per_km2 = 10**6
            area = gdf_drawn_albers.area.values[0] / m2_per_km2
            polyarea_label = 'Total area of satellite data to extract'
            polyarea_text = f"<b>{polyarea_label}</b>: {area:.2f} km<sup>2</sup>"

            # Test area size
            if self.max_size:
                confirmation_text = ('<span style="color: #33cc33"> '
                                     '<b>(Overriding maximum size limit; use with caution as may lead to memory issues)</b></span>')
                self.header.value = header_title_text + instruction_text + polyarea_text + confirmation_text
                self.gdf_drawn = gdf                
            elif area <= 10000:
                confirmation_text = ('<span style="color: #33cc33"> '
                                     '<b>(Area to extract falls within '
                                     'recommended limit)</b></span>')
                self.header.value = header_title_text + instruction_text + polyarea_text + confirmation_text
                self.gdf_drawn = gdf
            else:
                warning_text = ('<span style="color: #ff5050"> '
                                '<b>(Area to extract is too large, '
                                'please select an area less than 10000 '
                                'km<sup>2)</b></span>')
                self.header.value = header_title_text + instruction_text + polyarea_text + warning_text
                self.gdf_drawn = None

        ###########################
        # WIDGETS FOR APP OUTPUTS #
        ###########################

        self.status_info = Output(layout=make_box_layout())
        self.output_plot = Output(layout=make_box_layout())

        #########################################
        # MAP WIDGET, DRAWING TOOLS, WMS LAYERS #
        #########################################

        # Create drawing tools
        desired_drawtools = ['rectangle']
        draw_control = deawidgets.create_drawcontrol(desired_drawtools)

        # Begin by displaying an empty layer group, and update the group with desired WMS on interaction.
        self.map_layers = LayerGroup(layers=())
        self.map_layers.name = 'Map Overlays'

        # Create map widget
        self.m = deawidgets.create_map(map_center=(-28, 135), zoom_level=4)
        self.m.layout = make_box_layout()

        # Add tools to map widget
        self.m.add_control(draw_control)
        self.m.add_layer(self.map_layers)

        # Update all maps to starting defaults
        update_map_layers(self)

        ############################
        # WIDGETS FOR APP CONTROLS #
        ############################

        # Create parameter widgets
        dropdown_basemap = deawidgets.create_dropdown(self.basemap_list,
                                                      self.basemap_list[0][1])
        dropdown_dealayer = deawidgets.create_dropdown(self.dealayer_list,
                                                       self.dealayer_list[0][1])
        dropdown_output = deawidgets.create_dropdown(self.output_list,
                                                     self.output_list[0][1])
        date_picker = deawidgets.create_datepicker(
            value=datetime.date(2021, 12, 31))
        dropdown_styles = deawidgets.create_dropdown(self.styles_list,
                                                     self.styles_list[0])
        slider_abs = widgets.IntRangeSlider(value=[50, 3000],
                                            min=0,
                                            max=10000,
                                            step=25,
                                            description='',
                                            layout={'width': '85%'})
        run_button = create_expanded_button("Export imagery", "info")

        # Expandable advanced section
        text_resolution = widgets.FloatText(value=30,
                                            description='',
                                            layout={'width': '100%',
                                                    'margin': '0px',
                                                    'padding': '0px'})
        checkbox_pansharpen = deawidgets.create_checkbox(
            self.pansharpen, 'Pansharpen Landsat')
        slider_power = widgets.FloatSlider(value=1.0,
                                           min=0.01,
                                           max=1.0,
                                           step=0.01,
                                           description='',
                                            layout={'width': '85%'})
        checkbox_unsharp_mask = deawidgets.create_checkbox(
            self.unsharp_mask, 'Enable', layout={'width': '100%'})
        text_unsharp_mask_radius = widgets.FloatText(value=20,
                                                     step=1,
                                                     description='Radius',
                                                     layout={'width': '100%',
                                                             'margin': '0px',
                                                             'padding': '0px',
                                                             'display': 'none'})
        text_unsharp_mask_amount = widgets.FloatText(value=0.3,
                                                     step=0.1,
                                                     description='Amount',
                                                     layout={'width': '100%',
                                                             'margin': '0px',
                                                             'padding': '0px',
                                                             'display': 'none'})
        checkbox_max_size = deawidgets.create_checkbox(
            self.max_size, 'Enable')        
        text_dpi = widgets.IntText(value=0, 
                                     description="", 
                                     step=50, 
                                     layout={'width': '85%'})
        html_dpi = HTML('</br>Minimum DPI for image export</br>(100 DPI = 1000 pixels wide):')        
        expand_box = widgets.VBox([HTML('Resolution (metres):'),
                                   text_resolution,
                                   checkbox_pansharpen,
                                   HTML('</br>Apply power transformation to darken bright features:'),
                                   slider_power,
                                   HTML('</br>Apply unsharp masking to sharpen image:'),
                                   checkbox_unsharp_mask,
                                   text_unsharp_mask_radius,
                                   text_unsharp_mask_amount,
                                   HTML('</br>Override maximum size limit: (use with caution; may cause memory issues/crashes)'),
                                   checkbox_max_size,
                                   html_dpi,
                                   text_dpi],
                                  layout={'overflow': 'hidden'},
                                 )

        expand = widgets.Accordion(children=[expand_box], selected_index=None)
        expand.set_title(0, 'Advanced')

        # Add specific dialogs to class so they can be modified
        self.text_resolution = text_resolution
        self.checkbox_pansharpen = checkbox_pansharpen
        self.text_unsharp_mask_radius = text_unsharp_mask_radius
        self.text_unsharp_mask_amount = text_unsharp_mask_amount
        self.html_dpi = html_dpi

        ####################################
        # UPDATE FUNCTIONS FOR EACH WIDGET #
        ####################################

        # Run update functions whenever various widgets are changed.
        date_picker.observe(self.update_date, "value")
        dropdown_basemap.observe(self.update_basemap, "value")
        dropdown_dealayer.observe(self.update_dealayer, "value")
        dropdown_styles.observe(self.update_styles, "value")
        dropdown_output.observe(self.update_output, "value")
        run_button.on_click(self.run_app)
        draw_control.on_draw(update_geojson)
        slider_abs.observe(self.update_slider_abs, "value")
        
        # Advanced params
        text_resolution.observe(self.update_text_resolution, "value")
        checkbox_pansharpen.observe(self.update_checkbox_pansharpen, "value")
        slider_power.observe(self.update_slider_power, "value")
        checkbox_unsharp_mask.observe(self.update_checkbox_unsharp_mask, "value")
        text_unsharp_mask_radius.observe(self.update_text_unsharp_mask_radius, "value")
        text_unsharp_mask_amount.observe(self.update_text_unsharp_mask_amount, "value")
        checkbox_max_size.observe(self.update_checkbox_max_size, "value")
        text_dpi.observe(self.update_dpi, "value")
        

        ##################################
        # COLLECTION OF ALL APP CONTROLS #
        ##################################

        parameter_selection = VBox([
            HTML("<b>Date:</b>"),
            date_picker,
            HTML('<b>Satellite imagery:</b>'),
            dropdown_dealayer,
            HTML('<b>Style:</b>'),
            dropdown_styles,
            HTML('<b>Colour stretch:</b>'),
            slider_abs,
            HTML('<b>Output file format:</b>'),
            dropdown_output,
            HTML('</br>'),
            expand,
        ])
        map_selection = VBox([
            HTML("</br><b>Map overlay:</b>"),
            dropdown_basemap,
        ])
        parameter_selection.layout = make_box_layout()
        map_selection.layout = make_box_layout()

        ###############################
        # SPECIFICATION OF APP LAYOUT #
        ###############################

        #       0   1    2   3   4   5   6   7    8   9
        #     ---------------------------------------------
        # 0   | Header                         | Map sel. |
        #     |-------------------------------------------|
        # 1   | Params |                                  |
        # 2   |        |                                  |
        # 3   |        |                                  |
        # 4   |        |               Map                |
        # 5   |        |                                  |
        #     |--------|                                  |
        # 6   |  Run   |                                  |
        #     |-------------------------------------------|
        # 7   |   Status info   |      Figure/output      |
        # 8   |                 |                         |
        # 9   |                 |                         |
        # 10  |                 |                         |
        # 11  ---------------------------------------------

        # Create the layout #[rowspan, colspan]
        grid = GridspecLayout(12, 10, height="1400px", width="auto")

        # Header and controls
        grid[0, :8] = self.header
        grid[0, 8:] = map_selection
        grid[1:6, 0:2] = parameter_selection
        grid[6, 0:2] = run_button

        # Status info, map and plot
        grid[1:7, 2:] = self.m  # map
        grid[7:, 0:4] = self.status_info
        grid[7:, 4:] = self.output_plot

        # Display using HBox children attribute
        self.children = [grid]

    ######################################
    # DEFINITION OF ALL UPDATE FUNCTIONS #
    ######################################

    # Update date
    def update_date(self, change):
        self.date = str(change.new)
        update_map_layers(self)

    # Update colour stretch
    def update_slider_abs(self, change):
        self.vmin, self.vmax = change.new
    
    # Update power transform
    def update_slider_power(self, change):
        self.power = change.new

    # Enable pansharpening and reset/deactivate resolution
    def update_checkbox_pansharpen(self, change):
        self.pansharpen = change.new

        # Override default resolution if pansharpening is specified;
        # disable input if so
        if change.new:
            self.text_resolution.value = 15
            self.text_resolution.disabled = True
        else:
            self.text_resolution.value = 30
            self.text_resolution.disabled = False
            
    # Enable unsharp masking and show/hide custom params
    def update_checkbox_unsharp_mask(self, change):
        self.unsharp_mask = change.new
        
        # Show unsharp masking params in menu if activated
        if change.new:
            self.text_unsharp_mask_radius.layout.display = 'block'
            self.text_unsharp_mask_amount.layout.display = 'block'
        else:
            self.text_unsharp_mask_radius.layout.display = 'none'
            self.text_unsharp_mask_amount.layout.display = 'none'
            
    # Change unsharp masking radius
    def update_text_unsharp_mask_radius(self, change):
        self.unsharp_mask_radius = change.new
        
    # Change unsharp masking amount
    def update_text_unsharp_mask_amount(self, change):
        self.unsharp_mask_amount = change.new
        
    # Override max size limit
    def update_checkbox_max_size(self, change):
        self.max_size = change.new
        
    # Override min DPI
    def update_dpi(self, change):
        self.dpi = change.new
        
        # Update DPI helper text to give output resolution
        self.html_dpi.value = (f'</br>Minimum DPI for image export</br>'
                               f'({change.new} DPI = {change.new * 10} '
                               f'pixels wide):')

    # Update resolution
    def update_text_resolution(self, change):
        self.resolution = change.new

        # Clear data load params to trigger data re-load
        self.rgb_array = None
        self.sensor = None
        self.load_params = None
        self.query_params = None

    # Change layers shown on the map
    def update_dealayer(self, change):
        self.dealayer = change.new

        if change.new == 'ga_ls_ard_3':
            self.text_resolution.value = 30
            self.checkbox_pansharpen.disabled = False

            if self.pansharpen:
                self.text_resolution.value = 15
                self.text_resolution.disabled = True

        else:
            self.text_resolution.value = 10
            self.checkbox_pansharpen.disabled = True
            self.text_resolution.disabled = False

        update_map_layers(self)

    # Update basemap
    def update_basemap(self, change):
        self.basemap = change.new
        update_map_layers(self)

    # Set imagery style
    def update_styles(self, change):
        self.style = change.new
        update_map_layers(self)

    # Set output file format
    def update_output(self, change):
        self.output_format = change.new

    def run_app(self, change):

        # Clear progress bar and output areas before running
        self.status_info.clear_output()
        self.output_plot.clear_output()
        
        # Verify that polygon was drawn
        if self.gdf_drawn is not None:

            with self.status_info:

                    # Load data and add to attribute
                    if self.rgb_array is None:
                        self.rgb_array = extract_data(self)                            
                    else:
                        print('Using previously loaded data')                        

            if self.rgb_array is not None:
                
                with self.status_info:

                    # Create unique file name
                    centre_coords = self.gdf_drawn.geometry[0].centroid.coords[0][::-1]
                    site = reverse_geocode(coords=centre_coords)
                    fname = (f"{self.sensor} - {self.date} - {site} - {self.style}, "
                             f"{self.resolution:.0f} m resolution.{self.output_format}")

                    # Remove spaces and commas if requested
                    if self.standardise_name:
                        fname = (fname
                                 .replace(' - ', '_')
                                 .replace(', ', '-')
                                 .replace(' ', '-').lower())

                    print(f'\nExporting image for {site}.\nThis may take several minutes...')

                ############
                # Plotting #
                ############

                with self.output_plot:            
                    plot_data(self, fname)
            
            else:
                with self.status_info:
                    print('No satellite data found in the selected area. '
                          'Please select a new rectangle over an area with '
                          'satellite imagery visible on the map.')
        
        else:
            with self.status_info:
                print('Please draw a valid rectangle on the map, then press "Export imagery"')



# notebookapp_changefilmstrips.py
'''
This file contains functions for loading and interacting with data in the
change filmstrips notebook, inside the Real_world_examples folder.

Available functions:
    run_filmstrip_app

Last modified: September 2021
'''

# Load modules
import os
import dask
import datacube
import numpy as np
import pandas as pd
import xarray as xr
import matplotlib.pyplot as plt
from odc.algo import xr_geomedian
from odc.ui import select_on_a_map
from dask.utils import parse_bytes
from datacube.utils.geometry import CRS
from datacube.utils.rio import configure_s3_access
from datacube.utils.dask import start_local_dask
from ipyleaflet import basemaps, basemap_to_tiles

# Load utility functions
import sys
sys.path.insert(1, '../Tools/')
from dea_tools.datahandling import load_ard, mostcommon_crs
from dea_tools.coastal import tidal_tag
from dea_tools.dask import create_local_dask_cluster


def run_filmstrip_app(output_name,
                      time_range,
                      time_step,
                      tide_range=(0.0, 1.0),
                      resolution=(-30, 30),
                      max_cloud=50,
                      ls7_slc_off=False,
                      size_limit=200):
    '''
    An interactive app that allows the user to select a region from a
    map, then load Digital Earth Australia Landsat data and combine it
    using the geometric median ("geomedian") statistic to reveal the 
    median or 'typical' appearance of the landscape for a series of 
    time periods.
    
    The results for each time period are combined into a 'filmstrip' 
    plot which visualises how the landscape has changed in appearance 
    across time, with a 'change heatmap' panel highlighting potential 
    areas of greatest change.
    
    For coastal applications, the analysis can be customised to select 
    only satellite images obtained during a specific tidal range 
    (e.g. low, average or high tide).
    
    Last modified: June 2020

    Parameters
    ----------  
    output_name : str
        A name that will be used to name the output filmstrip plot file.
    time_range : tuple
        A tuple giving the date range to analyse 
        (e.g. `time_range = ('1988-01-01', '2017-12-31')`).
    time_step : dict
        This parameter sets the length of the time periods to compare 
        (e.g. `time_step = {'years': 5}` will generate one filmstrip 
        plot for every five years of data; `time_step = {'months': 18}` 
        will generate one plot for each 18 month period etc. Time 
        periods are counted from the first value given in `time_range`.
    tide_range : tuple, optional
        An optional parameter that can be used to generate filmstrip 
        plots based on specific ocean tide conditions. This can be 
        valuable for analysing change consistently along the coast. 
        For example, `tide_range = (0.0, 0.2)` will select only 
        satellite images acquired at the lowest 20% of tides; 
        `tide_range = (0.8, 1.0)` will select images from the highest 
        20% of tides. The default is `tide_range = (0.0, 1.0)` which 
        will select all images regardless of tide.
    resolution : tuple, optional
        The spatial resolution to load data. The default is 
        `resolution = (-30, 30)`, which will load data at 30 m pixel 
        resolution. Increasing this (e.g. to `resolution = (-100, 100)`) 
        can be useful for loading large spatial extents.
    max_cloud : int, optional
        This parameter can be used to exclude satellite images with 
        excessive cloud. The default is `50`, which will keep all images 
        with less than 50% cloud.
    ls7_slc_off : bool, optional
        An optional boolean indicating whether to include data from 
        after the Landsat 7 SLC failure (i.e. SLC-off). Defaults to 
        False, which removes all Landsat 7 observations > May 31 2003.
    size_limit : int, optional
        An optional size limit for the area selection in sq km.
        Defaults to 200 sq km.
        
    Returns
    -------
    ds_geomedian : xarray Dataset
        An xarray dataset containing geomedian composites for each 
        timestep in the analysis.
        
    '''    
    
    ########################
    # Select and load data #
    ########################
    
    # Define centre_coords as a global variable
    global centre_coords
        
    # Test if centre_coords is in the global namespace;
    # use default value if it isn't
    if 'centre_coords' not in globals():
        centre_coords = (-33.9719, 151.1934)
    
    # Plot interactive map to select area
    basemap = basemap_to_tiles(basemaps.Esri.WorldImagery)
    geopolygon = select_on_a_map(height='600px',
                                 layers=(basemap,),
                                 center=centre_coords , zoom=12)
        
    # Set centre coords based on most recent selection to re-focus
    # subsequent data selections
    centre_coords = geopolygon.centroid.points[0][::-1]

    # Test size of selected area
    area = geopolygon.to_crs(crs=CRS('epsg:3577')).area / 1000000
    if area > size_limit: 
        print(f'Warning: Your selected area is {area:.00f} sq km. '
              f'Please select an area of less than {size_limit} sq km.'
              f'\nTo select a smaller area, re-run the cell '
              f'above and draw a new polygon.')
        
    else:
        
        print('Starting analysis...')
        
        # Connect to datacube database
        dc = datacube.Datacube(app='Change_filmstrips')   
        
        # Configure local dask cluster
        create_local_dask_cluster()
        
        # Obtain native CRS 
        crs = mostcommon_crs(dc=dc, 
                             product='ga_ls5t_ard_3', 
                             query={'time': '1990', 
                                    'geopolygon': geopolygon})
        
        # Create query based on time range, area selected, custom params
        query = {'time': time_range,
                 'geopolygon': geopolygon,
                 'output_crs': crs,
                 'gqa_iterative_mean_xy': [0, 1],
                 'cloud_cover': [0, max_cloud],
                 'resolution': resolution,
                 'dask_chunks': {'time': 1, 'x': 2000, 'y': 2000},
                 'align': (resolution[1] / 2.0, resolution[1] / 2.0)}

        # Load data from all three Landsats
        ds = load_ard(dc=dc, 
                      measurements=['nbart_red', 
                                    'nbart_green', 
                                    'nbart_blue'],  
                      products=['ga_ls5t_ard_3', 
                                'ga_ls7e_ard_3', 
                                'ga_ls8c_ard_3'], 
                      min_gooddata=0.0,
                      ls7_slc_off=ls7_slc_off,
                      **query)
        
        # Optionally calculate tides for each timestep in the satellite 
        # dataset and drop any observations out side this range
        if tide_range != (0.0, 1.0):
            ds = tidal_tag(ds=ds, tidepost_lat=None, tidepost_lon=None)
            min_tide, max_tide = ds.tide_height.quantile(tide_range).values
            ds = ds.sel(time = (ds.tide_height >= min_tide) & 
                               (ds.tide_height <= max_tide))
            ds = ds.drop('tide_height')
            print(f'    Keeping {len(ds.time)} observations with tides '
                  f'between {min_tide:.2f} and {max_tide:.2f} m')
        
        # Create time step ranges to generate filmstrips from
        bins_dt = pd.date_range(start=time_range[0], 
                                end=time_range[1], 
                                freq=pd.DateOffset(**time_step))

        # Bin all satellite observations by timestep. If some observations
        # fall outside the upper bin, label these with the highest bin
        labels = bins_dt.astype('str')
        time_steps = (pd.cut(ds.time.values, bins_dt, labels = labels[:-1])
                      .add_categories(labels[-1])
                      .fillna(labels[-1])) 
        time_steps_var = xr.DataArray(time_steps, coords=[ds.time], 
                                      name='timestep')

        # Resample data temporally into time steps, and compute geomedians
        geomedian_ds = (ds.groupby(time_steps_var)
                        .apply(lambda ds_subset:
                               xr_geomedian(ds_subset,
                                            num_threads=1,
                                            eps=0.2 * (1 / 10_000),
                                            nocheck=True)))

        print('\nGenerating geomedian composites and plotting '
              'filmstrips... (click the Dashboard link above for status)')
        geomedian_ds = geomedian_ds.compute()

        # Reset CRS that is lost during geomedian compositing
        geomedian_ds.attrs['crs'] = ds.crs
        

        ############
        # Plotting #
        ############
        
        # Convert to array and extract vmin/vmax
        output_array = geomedian_ds[['nbart_red', 'nbart_green',
                                     'nbart_blue']].to_array()
        percentiles = output_array.quantile(q=(0.02, 0.98)).values

        # Compute heatmap by first taking the log of the array (so
        # change in dark areas can be identified), then computing
        # standard deviation between all timesteps
        heatmap_ds = (np.log(output_array)
                      .std(dim=['timestep'])
                      .mean(dim='variable'))
        heatmap_ds.attrs['crs'] = ds.crs

        # Create the plot with one subplot more than timesteps in the 
        # dataset. Figure width is set based on the number of subplots 
        # and aspect ratio
        n_obs = output_array.sizes['timestep']
        ratio = output_array.sizes['x'] / output_array.sizes['y']
        fig, axes = plt.subplots(1, n_obs + 1, 
                                 figsize=(5 * ratio * (n_obs + 1), 5))
        fig.subplots_adjust(wspace=0.05, hspace=0.05)

        # If 'spatial_ref' coord exists, drop it before plotting
        if 'spatial_ref' in output_array.coords:
            output_array = output_array.drop('spatial_ref')

        # Add timesteps to the plot, set aspect to equal to preserve shape
        for i, ax_i in enumerate(axes.flatten()[:n_obs]):
            output_array.isel(timestep=i).plot.imshow(ax=ax_i,
                                                      vmin=percentiles[0],
                                                      vmax=percentiles[1])
            ax_i.get_xaxis().set_visible(False)
            ax_i.get_yaxis().set_visible(False)
            ax_i.set_aspect('equal')

        # Add change heatmap panel to final subplot
        heatmap_ds.plot.imshow(ax=axes.flatten()[-1],
                               robust=True,
                               cmap='magma',
                               add_colorbar=False)
        axes.flatten()[-1].get_xaxis().set_visible(False)
        axes.flatten()[-1].get_yaxis().set_visible(False)
        axes.flatten()[-1].set_aspect('equal')
        axes.flatten()[-1].set_title('Change heatmap')

        # Export to file
        date_string = '_'.join(time_range)
        ts_v = list(time_step.values())[0]
        ts_k = list(time_step.keys())[0]
        fig.savefig(f'filmstrip_{output_name}_{date_string}_{ts_v}{ts_k}.png',
                    dpi=150,
                    bbox_inches='tight',
                    pad_inches=0.1)

        return geomedian_ds, heatmap_ds



# notebookapp_miningrehab.py
'''
This file contains functions for loading and interacting with data in the
mining rehabilitation notebook, inside the Real_world_examples folder.

Available functions:
    load_miningrehab_data
    run_miningrehab_app

Last modified: September 2021
'''

# Load modules
from ipyleaflet import Map, GeoJSON, DrawControl, basemaps
import datetime as dt
import datacube
import geopandas as gpd
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import rasterio.features
import IPython
from IPython.display import display
import warnings
import ipywidgets as widgets
from datacube.utils import masking

# Load utility functions
import sys
sys.path.insert(1, '../Tools/')
from dea_tools.datahandling import wofs_fuser
from dea_tools.spatial import transform_geojson_wgs_to_epsg


def load_miningrehab_data():
    """
    Loads DEA Fractional Cover and Water Observations products for the mining
    case-study area.
    Last modified: July 2021

    outputs
    ds - data set containing masked Fractional Cover data from Landsat 8
    Masked values are set to 'nan'
    """
    
    # Suppress warnings
    warnings.filterwarnings("ignore")

    # Initialise the data cube. 'app' argument is used to identify this app
    dc = datacube.Datacube(app="mining-app")

    # Specify latitude and longitude ranges
    latitude = (-34.426512, -34.434517)
    longitude = (116.648123, 116.630731)

    # Specify the date range
    time = ("2015-06-01", "2018-06-30")

    # Construct the data cube query
    query = {
        "x": longitude,
        "y": latitude,
        "time": time,
        "output_crs": "EPSG:3577",
        "resolution": (-30, 30),
    }

    print("Loading DEA Fractional Cover")
    dataset_fc = dc.load(product="ga_ls_fc_3",
                         group_by='solar_day',
                         platform='landsat-8',
                         cloud_cover= (0, 10),
                         **query)

    print("Loading DEA Water Observations")
    dataset_wo = dc.load(product="ga_ls_wo_3",
                         group_by='solar_day',
                         platform='landsat-8',
                         fuse_func=wofs_fuser,
                         cloud_cover= (0, 10),
                         like=dataset_fc)

    # Match the data
    shared_times = np.intersect1d(dataset_fc.time, dataset_wo.time)
    ds_fc_matched = dataset_fc.sel(time=shared_times)
    ds_wo_matched = dataset_wo.sel(time=shared_times)

    # Mask FC
    dry_mask = masking.make_mask(ds_wo_matched, dry=True)

    # Get fractional masked fc dataset (as proportion of 1, rather than 100)
    ds_fc_masked = ds_fc_matched.where(dry_mask.water) / 100

    # Resample
    ds_resampled = ds_fc_masked  #.resample(time="3M").median()
    ds_resampled.attrs["crs"] = dataset_fc.crs

    # Return the data
    return ds_fc_masked


def run_miningrehab_app(ds):
    """
    Plots an interactive map of the mining case-study area and allows
    the user to draw polygons. This returns plots of the fractional cover value
    of bare soil, green vegetation and brown vegetation in the polygon area.
    Last modified: January 2020

    inputs
    ds - data set containing masked Fractional Cover data from Landsat 8
    """
    
    # Suppress warnings
    warnings.filterwarnings("ignore")

    # Update plotting functionality through rcParams
    mpl.rcParams.update({"figure.autolayout": True})

    # Define the bounding box that will be overlayed on the interactive map
    # The bounds are hard-coded to match those from the loaded data
    geom_obj = {
        "type": "Feature",
        "properties": {
            "style": {
                "stroke": True,
                "color": "red",
                "weight": 4,
                "opacity": 0.8,
                "fill": True,
                "fillColor": False,
                "fillOpacity": 0,
                "showArea": True,
                "clickable": True,
            }
        },
        "geometry": {
            "type": "Polygon",
            "coordinates": [
                [
                    [116.630731, -34.434517],
                    [116.630731, -34.426512],
                    [116.648123, -34.426512],
                    [116.648123, -34.434517],
                    [116.630731, -34.434517],
                ]
            ],
        },
    }

    # Create a map geometry from the geom_obj dictionary
    # center specifies where the background map view should focus on
    # zoom specifies how zoomed in the background map should be
    centroid = gpd.GeoDataFrame.from_features([geom_obj]).geometry.centroid
    loadeddata_center = centroid.y.item(), centroid.x.item()
    loadeddata_zoom = 15

    # define the study area map
    studyarea_map = Map(
        layout=widgets.Layout(width="480px", height="600px"),
        center=loadeddata_center,
        zoom=loadeddata_zoom,
        basemap=basemaps.Esri.WorldImagery,
    )

    # define the drawing controls
    studyarea_drawctrl = DrawControl(
        polygon={"shapeOptions": {"fillOpacity": 0}},
        marker={},
        circle={},
        circlemarker={},
        polyline={},
    )

    # add drawing controls and data bound geometry to the map
    studyarea_map.add_control(studyarea_drawctrl)
    studyarea_map.add_layer(GeoJSON(data=geom_obj))

    # Index to count drawn polygons
    polygon_number = 0

    # Define widgets to interact with
    instruction = widgets.Output(layout={"border": "1px solid black"})
    with instruction:
        print(
            "Draw a polygon within the red box to view plots of "
            "the fractional cover values of bare, green and "
            "non-green cover for the area over time."
        )

    info = widgets.Output(layout={"border": "1px solid black"})
    with info:
        print("Plot status:")

    fig_display = widgets.Output(
        layout=widgets.Layout(
            width="50%"  # proportion of horizontal space taken by plot
        )
    )

    with fig_display:
        plt.ioff()
        fig, ax = plt.subplots(3, 1, figsize=(9, 9))

        for axis in ax:
            axis.set_ylim([0, 1])

    colour_list = plt.rcParams["axes.prop_cycle"].by_key()["color"]

    # Function to execute each time something is drawn on the map
    def handle_draw(self, action, geo_json):
        nonlocal polygon_number

        #         info.clear_output(wait=True)  # wait=True reduces flicker effect
        #         with info:
        #             print("Plot status: entered handle draw")

        # Execute behaviour based on what the user draws
        if geo_json["geometry"]["type"] == "Polygon":

            # Convert the drawn geometry to pixel coordinates
            geom_selectedarea = transform_geojson_wgs_to_epsg(
                geo_json,
                EPSG=3577  # hard-coded to be same as case-study data
            )

            # Construct a mask to only select pixels within the drawn polygon
            mask = rasterio.features.geometry_mask(
                [geom_selectedarea for geoms in [geom_selectedarea]],
                out_shape=ds.geobox.shape,
                transform=ds.geobox.affine,
                all_touched=False,
                invert=True,
            )

            masked_ds = ds.where(mask)
            masked_ds_mean = masked_ds.mean(dim=["x", "y"], skipna=True)

            colour = colour_list[polygon_number % len(colour_list)]

            # Add a layer to the map to make the most recently drawn polygon
            # the same colour as the line on the plot
            studyarea_map.add_layer(
                GeoJSON(
                    data=geo_json,
                    style={
                        "color": colour,
                        "opacity": 1,
                        "weight": 4.5,
                        "fillOpacity": 0.0,
                    },
                )
            )

            # Add Fractional cover plots to app
            masked_ds_mean = masked_ds_mean.drop('spatial_ref')
            masked_ds_mean.bs.interpolate_na(dim="time", method="nearest").plot.line("-", ax=ax[0])
            masked_ds_mean.pv.interpolate_na(dim="time", method="nearest").plot.line("-", ax=ax[1])
            masked_ds_mean.npv.interpolate_na(dim="time", method="nearest").plot.line("-", ax=ax[2])

            # reset titles back to custom
            ax[0].set_ylabel("Bare soil")
            ax[1].set_ylabel("Green vegetation")
            ax[2].set_ylabel("Brown (i.e. non-green) vegetation")
            ax[0].set_xlabel("")
            ax[1].set_xlabel("")
            ax[2].set_xlabel(None)

            # refresh display
            fig_display.clear_output(wait=True)  # wait=True reduces flicker effect
            with fig_display:
                display(fig)

            # Update plot info
            info.clear_output(wait=True)  # wait=True reduces flicker effect
            with info:
                print("Plot status: polygon added to plot")

            # Iterate the polygon number before drawing another polygon
            polygon_number = polygon_number + 1

        else:
            info.clear_output(wait=True)
            with info:
                print(
                    "Plot status: this drawing tool is not currently "
                    "supported. Please use the polygon tool."
                )

    # call to say activate handle_draw function on draw
    studyarea_drawctrl.on_draw(handle_draw)

    with fig_display:
        # TODO: update with user friendly something
        display(widgets.HTML(""))

    # Construct UI:
    #  +-----------------------+
    #  | instruction           |
    #  +-----------+-----------+
    #  |  map      |  plot     |
    #  |           |           |
    #  +-----------+-----------+
    #  | info                  |
    #  +-----------------------+
    ui = widgets.VBox([instruction, widgets.HBox([studyarea_map, fig_display]), info])
    display(ui)


# -*- coding: utf-8 -*-
# Land_cover_plotting.py
"""
Description: This file contains a set of python functions for plotting 
Digital Earth Australia Land Cover data.
License: The code in this notebook is licensed under the Apache License, 
Version 2.0 (https://www.apache.org/licenses/LICENSE-2.0). Digital Earth 
Australia data is licensed under the Creative Commons by Attribution 4.0 
license (https://creativecommons.org/licenses/by/4.0/).
Contact: If you need assistance, please post a question on the Open Data
Cube Slack channel (http://slack.opendatacube.org/) or on the GIS Stack 
Exchange (https://gis.stackexchange.com/questions/ask?tags=open-data-cube) 
using the `open-data-cube` tag (you can view previously asked questions 
here: https://gis.stackexchange.com/questions/tagged/open-data-cube). 
If you would like to report an issue with this script, file one on 
Github: https://github.com/GeoscienceAustralia/dea-notebooks/issues/new
Functions included:
    get_layer_name
    lc_colourmap
    plot_land_cover
    lc_animation
Last modified: January 2022
"""

import numpy as np
import pandas as pd
import ast
import sys

from IPython.display import Image

import matplotlib.pyplot as plt
from matplotlib import colors as mcolours
from matplotlib import patheffects
from matplotlib.animation import FuncAnimation

# Define colour schemes for each land cover measurement
lc_colours = {
    'level3': {0: (255, 255, 255, 255, "No Data"),
               111: (172, 188, 45, 255, "Cultivated Terrestrial\n Vegetation"),
               112: (14, 121, 18, 255, "Natural Terrestrial\n Vegetation"),
               124: (30, 191, 121, 255, "Natural Aquatic\n Vegetation"),
               215: (218, 92, 105, 255, "Artificial Surface"),
               216: (243, 171, 105, 255, "Natural Bare\n Surface"),
               220: (77, 159, 220, 255, "Water")},

    'level3_change_colour_scheme': {0: (255, 255, 255, 255, "No Change"),
                                    111112: (14, 121, 18, 255, "CTV -> NTV"),
                                    111215: (218, 92, 105, 255, "CTV -> AS"),
                                    111216: (243, 171, 105, 255, "CTV -> BS"),
                                    111220: (77, 159, 220, 255, "CTV -> Water"),
                                    112111: (172, 188, 45, 255, "NTV -> CTV"),
                                    112215: (218, 92, 105, 255, "NTV -> AS"),
                                    112216: (243, 171, 105, 255, "NTV -> BS"),
                                    112220: (77, 159, 220, 255, "NTV -> Water"),
                                    124220: (77, 159, 220, 255, "NAV -> Water"),
                                    215111: (172, 188, 45, 255, "AS -> CTV"),
                                    215112: (14, 121, 18, 255, "AS -> NTV"),
                                    215216: (243, 171, 105, 255, "AS -> BS"),
                                    215220: (77, 159, 220, 255, "AS -> Water"),
                                    216111: (172, 188, 45, 255, "BS -> CTV"),
                                    216112: (14, 121, 18,  255, "BS -> NTV"),
                                    216215: (218, 92, 105, 255, "BS -> AS"),
                                    216220: (77, 159, 220, 255, "BS -> Water"),
                                    220112: (14, 121, 18, 255, "Water -> NTV"),
                                    220216: (243, 171, 105, 255, "Water -> BS")},

    'level3_change_colour_bar': {0: (255, 255, 255, 255, "No change"),
                                 111: (172, 188, 45, 255, "Changed to Cultivated\n Terrestrial Vegetation"),
                                 112: (14, 121, 18, 255, "Changed to Natural\n Terrestrial Vegetation"),
                                 124: (30, 191, 121, 255, "Changed to Natural\n Aquatic Vegetation"),
                                 215: (218, 92, 105, 255, "Changed to Artificial\n Surface"),
                                 216: (243, 171, 105, 255, "Changed to Natural\n Bare Surface"),
                                 220: (77, 159, 220, 255, "Changed to Water")},

    'lifeform_veg_cat_l4a': {0: (255, 255, 255, 255, "No Data /\n Not vegetated"),
                             1: (14, 121, 18, 255, "Woody Vegetation"),
                             2: (172, 188, 45, 255, "Herbaceous\n Vegetation")},

    'canopyco_veg_cat_l4d': {0: (255, 255, 255, 255, "No Data /\n Not vegetated"),
                             10: (14,  121, 18,  255, "> 65 % cover"),
                             12: (45,  141, 47,  255, "40 to 65 % cover"),
                             13: (80,  160, 82,  255, "15 to 40 % cover"),
                             15: (117, 180, 118, 255, "4 to 15 % cover"),
                             16: (154, 199, 156, 255, "1 to 4 % cover")},

    'waterstt_wat_cat_l4a': {0: (255, 255, 255, 255, "No Data /\n Not water"),
                             1: (77, 159, 220, 255, "Water")},

    'watersea_veg_cat_l4a_au': {0: (255, 255, 255, 255, "No data /\n Not aquatic vegetation"),
                                1: (25,  173, 109, 255, "> 3 months"),
                                2: (176, 218, 201, 255, "< 3 months")},

    'inttidal_wat_cat_l4a': {0: (255, 255, 255, 255, "No data /\n Not intertidal"),
                             3: (77, 159, 220, 255, "Intertidal")},

    'waterper_wat_cat_l4d_au': {0: (255, 255, 255, 255, "No data /\n Not water"),
                                1: (27,  85,  186, 255, "> 9 months"),
                                7: (52,  121, 201, 255, "7 to 9 months"),
                                8: (79,  157, 217, 255, "4 to 6 months"),
                                9: (113, 202, 253, 255, "1 to 3 months")},

    'baregrad_phy_cat_l4d_au': {0: (255, 255, 255, 255, "No data /\n Not bare"),
                                10: (255, 230, 140, 255, "Sparsely vegetated\n (< 20% bare)"),
                                12: (250, 210, 110, 255, "Very sparsely\n vegetated (20 to 60% bare)"),
                                15: (243, 171, 105, 255, "Bare areas,\n unvegetated (> 60% bare)")},

    'level4': {0: (255, 255, 255, 255, "No Data"),
               1: (151, 187, 26, 255, 'Cultivated Terrestrial\n Vegetated:'),
               2: (151, 187, 26, 255, 'Cultivated Terrestrial\n Vegetated: Woody'),
               3: (209, 224, 51, 255, 'Cultivated Terrestrial\n Vegetated: Herbaceous'),
               4: (197, 168, 71, 255, 'Cultivated Terrestrial\n Vegetated: Closed\n (> 65 %)'),
               5: (205, 181, 75, 255, 'Cultivated Terrestrial\n Vegetated: Open\n (40 to 65 %)'),
               6: (213, 193, 79, 255, 'Cultivated Terrestrial\n Vegetated: Open\n (15 to 40 %)'),
               7: (228, 210, 108, 255, 'Cultivated Terrestrial\n Vegetated: Sparse\n (4 to 15 %)'),
               8: (242, 227, 138, 255, 'Cultivated Terrestrial\n Vegetated: Scattered\n (1 to 4 %)'),
               # 9: (197, 168, 71, 255, 'Cultivated Terrestrial\n Vegetated: Woody Closed\n (> 65 %)'),
               # 10: (205, 181, 75, 255, 'Cultivated Terrestrial\n Vegetated: Woody Open\n (40 to 65 %)'),
               # 11: (213, 193, 79, 255, 'Cultivated Terrestrial\n Vegetated: Woody Open\n (15 to 40 %)'),
               # 12: (228, 210, 108, 255, 'Cultivated Terrestrial\n Vegetated: Woody Sparse\n (4 to 15 %)'),
               # 13: (242, 227, 138, 255, 'Cultivated Terrestrial\n Vegetated: Woody Scattered\n (1 to 4 %)'),
               14: (228, 224, 52, 255, 'Cultivated Terrestrial\n Vegetated: Herbaceous Closed\n (> 65 %)'),
               15: (235, 232, 84, 255, 'Cultivated Terrestrial\n Vegetated: Herbaceous Open\n (40 to 65 %)'),
               16: (242, 240, 127, 255, 'Cultivated Terrestrial\n Vegetated: Herbaceous Open\n (15 to 40 %)'),
               17: (249, 247, 174, 255, 'Cultivated Terrestrial\n Vegetated: Herbaceous Sparse\n (4 to 15 %)'),
               18: (255, 254, 222, 255, 'Cultivated Terrestrial\n Vegetated: Herbaceous Scattered\n (1 to 4 %)'),
               19: (14, 121, 18, 255, 'Natural Terrestrial Vegetated:'),
               20: (26, 177, 87, 255, 'Natural Terrestrial Vegetated: Woody'),
               21: (94, 179, 31, 255, 'Natural Terrestrial Vegetated: Herbaceous'),
               22: (14, 121, 18, 255, 'Natural Terrestrial Vegetated: Closed (> 65 %)'),
               23: (45, 141, 47, 255, 'Natural Terrestrial Vegetated: Open (40 to 65 %)'),
               24: (80, 160, 82, 255, 'Natural Terrestrial Vegetated: Open (15 to 40 %)'),
               25: (117, 180, 118, 255, 'Natural Terrestrial Vegetated: Sparse (4 to 15 %)'),
               26: (154, 199, 156, 255, 'Natural Terrestrial Vegetated: Scattered (1 to 4 %)'),
               27: (14, 121, 18, 255, 'Natural Terrestrial Vegetated: Woody Closed (> 65 %)'),
               28: (45, 141, 47, 255, 'Natural Terrestrial Vegetated: Woody Open (40 to 65 %)'),
               29: (80, 160, 82, 255, 'Natural Terrestrial Vegetated: Woody Open (15 to 40 %)'),
               30: (117, 180, 118, 255, 'Natural Terrestrial Vegetated: Woody Sparse (4 to 15 %)'),
               31: (154, 199, 156, 255, 'Natural Terrestrial Vegetated: Woody Scattered (1 to 4 %)'),
               32: (119, 167, 30, 255, 'Natural Terrestrial Vegetated: Herbaceous Closed (> 65 %)'),
               33: (136, 182, 51, 255, 'Natural Terrestrial Vegetated: Herbaceous Open (40 to 65 %)'),
               34: (153, 196, 80, 255, 'Natural Terrestrial Vegetated: Herbaceous Open (15 to 40 %)'),
               35: (170, 212, 113, 255, 'Natural Terrestrial Vegetated: Herbaceous Sparse (4 to 15 %)'),
               36: (186, 226, 146, 255, 'Natural Terrestrial Vegetated: Herbaceous Scattered (1 to 4 %)'),
               # 37: (86, 236, 231, 255, 'Cultivated Aquatic Vegetated:'),
               # 38: (61, 170, 140, 255, 'Cultivated Aquatic Vegetated: Woody'),
               # 39: (82, 231, 172, 255, 'Cultivated Aquatic Vegetated: Herbaceous'),
               # 40: (43, 210, 203, 255, 'Cultivated Aquatic Vegetated: Closed (> 65 %)'),
               # 41: (73, 222, 216, 255, 'Cultivated Aquatic Vegetated: Open (40 to 65 %)'),
               # 42: (110, 233, 228, 255, 'Cultivated Aquatic Vegetated: Open (15 to 40 %)'),
               # 43: (149, 244, 240, 255, 'Cultivated Aquatic Vegetated: Sparse (4 to 15 %)'),
               # 44: (187, 255, 252, 255, 'Cultivated Aquatic Vegetated: Scattered (1 to 4 %)'),
               # 45: (43, 210, 203, 255, 'Cultivated Aquatic Vegetated: Woody Closed (> 65 %)'),
               # 46: (73, 222, 216, 255, 'Cultivated Aquatic Vegetated: Woody Open (40 to 65 %)'),
               # 47: (110, 233, 228, 255, 'Cultivated Aquatic Vegetated: Woody Open (15 to 40 %)'),
               # 48: (149, 244, 240, 255, 'Cultivated Aquatic Vegetated: Woody Sparse (4 to 15 %)'),
               # 49: (187, 255, 252, 255, 'Cultivated Aquatic Vegetated: Woody Scattered (1 to 4 %)'),
               # 50: (82, 231, 196, 255, 'Cultivated Aquatic Vegetated: Herbaceous Closed (> 65 %)'),
               # 51: (113, 237, 208, 255, 'Cultivated Aquatic Vegetated: Herbaceous Open (40 to 65 %)'),
               # 52: (144, 243, 220, 255, 'Cultivated Aquatic Vegetated: Herbaceous Open (15 to 40 %)'),
               # 53: (175, 249, 232, 255, 'Cultivated Aquatic Vegetated: Herbaceous Sparse (4 to 15 %)'),
               # 54: (207, 255, 244, 255, 'Cultivated Aquatic Vegetated: Herbaceous Scattered (1 to 4 %)'),
               55: (30, 191, 121, 255, 'Natural Aquatic Vegetated:'),
               56: (18, 142, 148, 255, 'Natural Aquatic Vegetated: Woody'),
               57: (112, 234, 134, 255, 'Natural Aquatic Vegetated: Herbaceous'),
               58: (25, 173, 109, 255, 'Natural Aquatic Vegetated: Closed (> 65 %)'),
               59: (53, 184, 132, 255, 'Natural Aquatic Vegetated: Open (40 to 65 %)'),
               60: (93, 195, 155, 255, 'Natural Aquatic Vegetated: Open (15 to 40 %)'),
               61: (135, 206, 178, 255, 'Natural Aquatic Vegetated: Sparse (4 to 15 %)'),
               62: (176, 218, 201, 255, 'Natural Aquatic Vegetated: Scattered (1 to 4 %)'),
               63: (25, 173, 109, 255, 'Natural Aquatic Vegetated: Woody Closed (> 65 %)'),
               64: (25, 173, 109, 255, 'Natural Aquatic Vegetated: Woody Closed (> 65 %) Water > 3 months (semi-) permenant'),
               65: (25, 173, 109, 255, 'Natural Aquatic Vegetated: Woody Closed (> 65 %) Water < 3 months (temporary or seasonal)'),
               66: (53, 184, 132, 255, 'Natural Aquatic Vegetated: Woody Open (40 to 65 %)'),
               67: (53, 184, 132, 255, 'Natural Aquatic Vegetated: Woody Open (40 to 65 %) Water > 3 months (semi-) permenant'),
               68: (53, 184, 132, 255, 'Natural Aquatic Vegetated: Woody Open (40 to 65 %) Water < 3 months (temporary or seasonal)'),
               69: (93, 195, 155, 255, 'Natural Aquatic Vegetated: Woody Open (15 to 40 %)'),
               70: (93, 195, 155, 255, 'Natural Aquatic Vegetated: Woody Open (15 to 40 %) Water > 3 months (semi-) permenant'),
               71: (93, 195, 155, 255, 'Natural Aquatic Vegetated: Woody Open (15 to 40 %) Water < 3 months (temporary or seasonal)'),
               72: (135, 206, 178, 255, 'Natural Aquatic Vegetated: Woody Sparse (4 to 15 %)'),
               73: (135, 206, 178, 255, 'Natural Aquatic Vegetated: Woody Sparse (4 to 15 %) Water > 3 months (semi-) permenant'),
               74: (135, 206, 178, 255, 'Natural Aquatic Vegetated: Woody Sparse (4 to 15 %) Water < 3 months (temporary or seasonal)'),
               75: (176, 218, 201, 255, 'Natural Aquatic Vegetated: Woody Scattered (1 to 4 %)'),
               76: (176, 218, 201, 255, 'Natural Aquatic Vegetated: Woody Scattered (1 to 4 %) Water > 3 months (semi-) permenant'),
               77: (176, 218, 201, 255, 'Natural Aquatic Vegetated: Woody Scattered (1 to 4 %) Water < 3 months (temporary or seasonal)'),
               78: (39, 204, 139, 255, 'Natural Aquatic Vegetated: Herbaceous Closed (> 65 %)'),
               79: (39, 204, 139, 255, 'Natural Aquatic Vegetated: Herbaceous Closed (> 65 %) Water > 3 months (semi-) permenant'),
               80: (39, 204, 139, 255, 'Natural Aquatic Vegetated: Herbaceous Closed (> 65 %) Water < 3 months (temporary or seasonal)'),
               81: (66, 216, 159, 255, 'Natural Aquatic Vegetated: Herbaceous Open (40 to 65 %)'),
               82: (66, 216, 159, 255, 'Natural Aquatic Vegetated: Herbaceous Open (40 to 65 %) Water > 3 months (semi-) permenant'),
               83: (66, 216, 159, 255, 'Natural Aquatic Vegetated: Herbaceous Open (40 to 65 %) Water < 3 months (temporary or seasonal)'),
               84: (99, 227, 180, 255, 'Natural Aquatic Vegetated: Herbaceous Open (15 to 40 %)'),
               85: (99, 227, 180, 255, 'Natural Aquatic Vegetated: Herbaceous Open (15 to 40 %) Water > 3 months (semi-) permenant'),
               86: (99, 227, 180, 255, 'Natural Aquatic Vegetated: Herbaceous Open (15 to 40 %) Water < 3 months (temporary or seasonal)'),
               87: (135, 239, 201, 255, 'Natural Aquatic Vegetated: Herbaceous Sparse (4 to 15 %)'),
               88: (135, 239, 201, 255, 'Natural Aquatic Vegetated: Herbaceous Sparse (4 to 15 %) Water > 3 months (semi-) permenant'),
               89: (135, 239, 201, 255, 'Natural Aquatic Vegetated: Herbaceous Sparse (4 to 15 %) Water < 3 months (temporary or seasonal)'),
               90: (171, 250, 221, 255, 'Natural Aquatic Vegetated: Herbaceous Scattered (1 to 4 %)'),
               91: (171, 250, 221, 255, 'Natural Aquatic Vegetated: Herbaceous Scattered (1 to 4 %) Water > 3 months (semi-) permenant'),
               92: (171, 250, 221, 255, 'Natural Aquatic Vegetated: Herbaceous Scattered (1 to 4 %) Water < 3 months (temporary or seasonal)'),
               93: (218, 92, 105, 255, 'Artificial Surface:'),
               94: (243, 171, 105, 255, 'Natural Surface:'),
               95: (255, 230, 140, 255, 'Natural Surface: Sparsely vegetated'),
               96: (250, 210, 110, 255, 'Natural Surface: Very sparsely vegetated'),
               97: (243, 171, 105, 255, 'Natural Surface: Bare areas, unvegetated'),
               98: (77, 159, 220, 255, 'Water:'),
               99: (77, 159, 220, 255, 'Water: (Water)'),
               100: (187, 220, 233, 255, 'Water: (Water) Tidal area'),
               101: (27, 85, 186, 255, 'Water: (Water) Perennial (> 9 months)'),
               102: (52, 121, 201, 255, 'Water: (Water) Non-perennial (7 to 9 months)'),
               103: (79, 157, 217, 255, 'Water: (Water) Non-perennial (4 to 6 months)'),
               104: (133, 202, 253, 255, 'Water: (Water) Non-perennial (1 to 3 months)'),
               # 105: (250, 250, 250, 255, 'Water: (Snow)')
               },

    'level4_colourbar_labels': {0: (255, 255, 255, 255, "No Data"),
                                14: (228, 224, 52, 255, 'Cultivated Terrestrial Vegetated: Herbaceous Closed (> 65 %)'),
                                15: (235, 232, 84, 255, 'Cultivated Terrestrial Vegetated: Herbaceous Open (40 to 65 %)'),
                                16: (242, 240, 127, 255, 'Cultivated Terrestrial Vegetated: Herbaceous Open (15 to 40 %)'),
                                17: (249, 247, 174, 255, 'Cultivated Terrestrial Vegetated: Herbaceous Sparse (4 to 15 %)'),
                                18: (255, 254, 222, 255, 'Cultivated Terrestrial Vegetated: Herbaceous Scattered (1 to 4 %)'),
                                27: (14, 121, 18, 255, 'Natural Terrestrial Vegetated: Woody Closed (> 65 %)'),
                                28: (45, 141, 47, 255, 'Natural Terrestrial Vegetated: Woody Open (40 to 65 %)'),
                                29: (80, 160, 82, 255, 'Natural Terrestrial Vegetated: Woody Open (15 to 40 %)'),
                                30: (117, 180, 118, 255, 'Natural Terrestrial Vegetated: Woody Sparse (4 to 15 %)'),
                                31: (154, 199, 156, 255, 'Natural Terrestrial Vegetated: Woody Scattered (1 to 4 %)'),
                                32: (119, 167, 30, 255, 'Natural Terrestrial Vegetated: Herbaceous Closed (> 65 %)'),
                                33: (136, 182, 51, 255, 'Natural Terrestrial Vegetated: Herbaceous Open (40 to 65 %)'),
                                34: (153, 196, 80, 255, 'Natural Terrestrial Vegetated: Herbaceous Open (15 to 40 %)'),
                                35: (170, 212, 113, 255, 'Natural Terrestrial Vegetated: Herbaceous Sparse (4 to 15 %)'),
                                36: (186, 226, 146, 255, 'Natural Terrestrial Vegetated: Herbaceous Scattered (1 to 4 %)'),
                                65: (25, 173, 109, 255, 'Natural Aquatic Vegetated: Woody Closed (> 65 %)'),
                                68: (53, 184, 132, 255, 'Natural Aquatic Vegetated: Woody Open (40 to 65 %)'),
                                71: (93, 195, 155, 255, 'Natural Aquatic Vegetated: Woody Open (15 to 40 %)'),
                                74: (135, 206, 178, 255, 'Natural Aquatic Vegetated: Woody Sparse (4 to 15 %)'),
                                77: (176, 218, 201, 255, 'Natural Aquatic Vegetated: Woody Scattered (1 to 4 %)'),
                                80: (39, 204, 139, 255, 'Natural Aquatic Vegetated: Herbaceous Closed (> 65 %)'),
                                83: (66, 216, 159, 255, 'Natural Aquatic Vegetated: Herbaceous Open (40 to 65 %)'),
                                86: (99, 227, 180, 255, 'Natural Aquatic Vegetated: Herbaceous Open (15 to 40 %)'),
                                89: (135, 239, 201, 255, 'Natural Aquatic Vegetated: Herbaceous Sparse (4 to 15 %)'),
                                92: (171, 250, 221, 255, 'Natural Aquatic Vegetated: Herbaceous Scattered (1 to 4 %)'),
                                93: (218, 92, 105, 255, 'Artificial Surface'),
                                95: (255, 230, 140, 255, 'Natural Surface: Sparsely vegetated'),
                                96: (250, 210, 110, 255, 'Natural Surface: Very sparsely vegetated'),
                                97: (243, 171, 105, 255, 'Natural Surface: Bare areas, unvegetated'),
                                100: (187, 220, 233, 255, 'Water: (Water) Tidal area'),
                                101: (27, 85, 186, 255, 'Water: (Water) Perennial (> 9 months)'),
                                102: (52, 121, 201, 255, 'Water: (Water) Non-perennial (7 to 9 months)'),
                                103: (79, 157, 217, 255, 'Water: (Water) Non-perennial (4 to 6 months)'),
                                104: (133, 202, 253, 255, 'Water: (Water) Non-perennial (1 to 3 months)')},
}


def get_layer_name(measurement, da):
    aliases = {
        'lifeform': 'lifeform_veg_cat_l4a',
        'vegetation_cover': 'canopyco_veg_cat_l4d',
        'water_seasonality': 'watersea_veg_cat_l4a_au',
        'water_state': 'waterstt_wat_cat_l4a',
        'intertidal': 'inttidal_wat_cat_l4a',
        'water_persistence': 'waterper_wat_cat_l4d_au',
        'bare_gradation': 'baregrad_phy_cat_l4d_au',
        'full_classification': 'level4',
        'level_4': 'level4'
    }

    # Use provided measurement if able
    measurement = measurement.lower() if measurement else da.name
    measurement = aliases[measurement] if measurement in aliases.keys(
    ) else measurement
    return measurement


def make_colorbar(fig, ax, measurement, horizontal=False, animation=False):
    """
    Adds a new colorbar with appropriate land cover colours and labels
    
    measurement = land cover measurement to use for colour map and lables
    
    for level 4, must be used with a double plot, ax should be the left side canvas
    the colour bar will go over thr right hand side.
    """
    # Create new axis object for colorbar
    # parameters for add_axes are [left, bottom, width, height], in
    # fractions of total plot
    
    if measurement == 'level4' and animation == True:
        
        # special spacing settings for level 4
        cax = fig.add_axes([0.62, 0.10, 0.02, 0.80])
        orient = 'vertical'
        
            # get level 4 colour bar colour map ect
        cb_cmap, cb_norm, cb_labels, cb_ticks = lc_colourmap('level4_colourbar_labels',
                                                         colour_bar=True)
    elif measurement == 'level4' and animation == False:
        
        # get level 4 colour bar colour map ect
        cb_cmap, cb_norm, cb_labels, cb_ticks = lc_colourmap('level4_colourbar_labels',
                                                         colour_bar=True)
        #move plot over to make room for colourbar
        fig.subplots_adjust(right=0.825)

        # Settings for axis positions
        cax = fig.add_axes([0.84, 0.15, 0.02, 0.70])
        orient = 'vertical'
        
    else:
        #for all other measurements 

        #move plot over to make room for colourbar
        fig.subplots_adjust(right=0.825)

        # Settings for different axis positions
        if horizontal:
            cax = fig.add_axes([0.02, 0.05, 0.90, 0.03])
            orient = 'horizontal'
        else:
            cax = fig.add_axes([0.84, 0.15, 0.02, 0.70])
            orient = 'vertical'
            
        # get measurement colour bar colour map ect
        cb_cmap, cb_norm, cb_labels, cb_ticks = lc_colourmap(measurement,
                                                         colour_bar=True)

    img = ax.imshow([cb_ticks], cmap=cb_cmap, norm=cb_norm)
    cb = fig.colorbar(img, cax=cax, orientation=orient)

    cb.ax.tick_params(labelsize=12)
    cb.set_ticks(cb_ticks + np.diff(cb_ticks, append=cb_ticks[-1]+1) / 2)
    cb.set_ticklabels(cb_labels)



def lc_colourmap(colour_scheme, colour_bar=False):
    """
    Returns colour map and normalisation for the provided DEA Land Cover
    measurement, for use in plotting with Matplotlib library
    Parameters
    ----------
    colour_scheme : string
        Name of land cover colour scheme to use
        Valid options: 'level3', 'level4', 'lifeform_veg_cat_l4a', 
        'canopyco_veg_cat_l4d', 'watersea_veg_cat_l4a_au',
        'waterstt_wat_cat_l4a', 'inttidal_wat_cat_l4a', 
        'waterper_wat_cat_l4d_au', 'baregrad_phy_cat_l4d_au'.
    colour_bar : bool, optional
        Controls if colour bar labels are returned as a list for 
        plotting a colour bar. Default: False.
    Returns
    ---------
    cmap : matplotlib colormap
        Matplotlib colormap containing the colour scheme for the
        specified DEA Land Cover measurement.
    norm : matplotlib colormap index
        Matplotlib colormap index based on the discrete intervals of the
        classes in the specified DEA Land Cover measurement. Ensures the
        colormap maps the colours to the class numbers correctly.
    cblables : array
        A two dimentional array containing the numerical class values
        (first dim) and string labels (second dim) of the classes found
        in the chosen DEA Land Cover measurement.
    """

    colour_scheme = colour_scheme.lower()
    # Ensure a valid colour scheme was requested
#     try:
    assert (colour_scheme in lc_colours.keys(
    )), f'colour scheme must be one of [{lc_colours.keys()}] (got "{colour_scheme}")'

#     ('The dataset provided does not have a valid '
#     'name. Please specify which DEA Landcover measurement is being plotted '
#     'by providing the name using the "measurement" variable. For example (measurement = "full_classification")')

    # Get colour definitions
    lc_colour_scheme = lc_colours[colour_scheme]

    # Create colour map
    colour_arr = []
    for key, value in lc_colour_scheme.items():
        colour_arr.append(np.array(value[:-2]) / 255)

    cmap = mcolours.ListedColormap(colour_arr)
    bounds = list(lc_colour_scheme)

    if colour_bar == True:
        if colour_scheme == 'level4':
            # Set colour labels to shortened level 4 list
            lc_colour_scheme = lc_colours['level4_colourbar_labels']
        cb_ticks = list(lc_colour_scheme)
        cb_labels = []
        for x in cb_ticks:
            cb_labels.append(lc_colour_scheme[x][4])

    bounds.append(bounds[-1]+1)
    norm = mcolours.BoundaryNorm(np.array(bounds), cmap.N)

    if colour_bar == False:
        return (cmap, norm)
    else:
        return (cmap, norm, cb_labels, cb_ticks)


def plot_land_cover(data, year=None, measurement=None, out_width=15, cols=4,):
    """
    Plot a single land cover measurement with appropriate colour scheme.
    Parameters
    ----------
    data : xarray.DataArray
        A dataArray containing a DEA Land Cover classification.
    year : int, optional
        Can be used to select to plot a specific year. If not provided,
        all time slices are plotted.
    measurement : string, optional
        Name of the DEA land cover classification to be plotted. Passed to 
        lc_colourmap to specify which colour scheme will ve used. If non 
        provided, reads data array name from `da` to determine.
    """
    # get measurement name
    measurement = get_layer_name(measurement, data)

    # get colour map, normalisation
    try:
        cmap, norm = lc_colourmap(measurement)
    except AssertionError:

        raise KeyError('Could not automatically determine colour scheme from'
                       f'DataArray name {measurement}. Please specify which '
                       'DEA Landcover measurement is being plotted by providing'
                       'the name using the "measurement" variable For example'
                       '(measurement = "full_classification")')

    height, width = data.geobox.shape
    scale = out_width / width

    if year:
        year_string = f"{year}-01-01"
        data = data.sel(time=year_string, method="nearest")

    # plot all dates for the provided measurement
    if len(data.dims) < 3:
        fig, ax = plt.subplots()
        fig.set_size_inches(width * scale, height * scale)
        make_colorbar(fig, ax, measurement)
        im = ax.imshow(data, cmap=cmap, norm=norm, interpolation="nearest")
    else:
        if cols > len(data.time):
            cols = len(data.time)
        rows = int((len(data.time) + cols-1)/cols)

        fig, ax = plt.subplots(nrows=rows, ncols=cols)
        fig.set_size_inches(
            width * scale, (height * scale / cols) * (len(data.time) / cols))

        make_colorbar(fig, ax.flat[0], measurement)

        for a, b in enumerate(ax.flat):
            if a < data.shape[0]:
                im = b.imshow(data[a], cmap=cmap, norm=norm,
                              interpolation="nearest")

    return im


def lc_animation(
        da,
        file_name="default_animation",
        measurement=None,
        stacked_plot=False,
        colour_bar=False,
        animation_interval=500,
        width_pixels=10,
        dpi=150,
        font_size=15,
        label_ax=True):
    """
    Creates an animation of DEA Landcover though time beside 
    corresponding stacked plots of the landcover classes. Saves the
    animation to a file and displays the animation in notebook.
    Inputs
    -------
    da : xarray.DataArray
        An xarray.DataArray containing a multi-date stack of 
        observations of a single landcover level.
    file_name: string, optional.
        string used to create filename for saved animation file.
        Default: "default_animation" code adds .gif suffix.
    measurement : string, optional
        Name of the DEA land cover classification to be plotted. Passed to 
        lc_colourmap to specify which colour scheme will ve used. If non 
        provided, reads data array name from `da` to determine.
    Stacked_plot: boolean, optional
        Determines if a stacked plot showing the percentage of area
        taken up by each class in each time slice is added to the
        animation. Default: False.
    colour_bar : boolean, Optional
        Determines if a colour bar is generated for the stand alone 
        animation. This is NOT recommended for use with level 4 data. 
        Does not work with stacked plot. Default: False.
    animation_interval : int , optional
        How quickly the frames of the animations should be re-drawn. 
        Default: 500.
    Width_pixels : int, optional
        How wide in pixles the animation plot should be. Default: 10.
    dpi : int, optional
        Stands for 'Dots Per Inch'. Passed to the fuction that saves the
        animation and determines the resolution. A higher number will
        produce a higher resolution image but a larger file size and
        slower processing. Default: 150.
    font_size : int, optional. 
        Controls the size of the text which indicates the year
        displayed. Default: 15.
    label_ax : boolean, optional
        Determines if animation plot should have tick marks and numbers
        on axes. Also removes white space around plot. default: True
    Returns
    ---------
    A .gif file animation.
    """

    def calc_class_ratio(da):
        """
        Creates a table listing year by year what percentage of the
        total area is taken up by each class.
        Parameters
        ----------
        da : xarray.DataArray with time dimension
        Returns
        -------
        Pandas Dataframe : containing class percentages per year
        """

        # list all class codes in dataset
        list_classes = (np.unique(da, return_counts=False)).tolist()

        # create empty dataframe & dictionary
        ratio_table = pd.DataFrame(data=None, columns=list_classes)
        date_line = {}

        # count all pixels, should be consistent
        total_pix = int(np.sum(da.isel(time=1)))

        # iterate through each year in dataset
        for i in range(0, len(da.time)):
            date = str(da.time[i].data)[0:10]

            # for each year iterate though each present class number
            # and count pixels
            for n in list_classes:
                number_of_pixles = int(np.sum(da.isel(time=i) == n))
                percentage = number_of_pixles / total_pix * 100
                date_line[n] = percentage

            # add each year's counts to dataframe
            ratio_table.loc[date] = date_line

        return ratio_table

    def rgb_to_hex(r, g, b):
        hex = "#%x%x%x" % (r, g, b)
        if len(hex) < 7:
            hex = "#0" + hex[1:]
        return hex

    measurement = get_layer_name(measurement, da)

    # Add gif to end of filename
    file_name = file_name + ".gif"

        # Create colour map and normalisation for specified lc measurement
    try:
        layer_cmap, layer_norm, cb_labels, cb_ticks = lc_colourmap(
            measurement, colour_bar=True)
    except AssertionError:

        raise KeyError(f'Could not automatically determine colour scheme from '
                   f'DataArray name {measurement}. Please specify which '
                   'DEA Landcover measurement is being plotted by providing '
                   'the name using the "measurement" variable For example '
                   '(measurement = "full_classification")')
    
    # Prepare variables needed
    # Get info on dataset dimensions
    height, width = da.geobox.shape
    scale = width_pixels / width
    left, bottom, right, top = da.geobox.extent.boundingbox
    extent = [left, right, bottom, top]

    outline = [patheffects.withStroke(linewidth=2.5, foreground="black")]
    annotation_defaults = {
        "xy": (1, 1),
        "xycoords": "axes fraction",
        "xytext": (-5, -5),
        "textcoords": "offset points",
        "horizontalalignment": "right",
        "verticalalignment": "top",
        "fontsize": font_size,
        "color": "white",
        "path_effects": outline,
    }

    # Get information needed to display the year in the top corner
    times_list = da.time.dt.strftime("%Y").values
    text_list = [False] * len(times_list)
    annotation_list = ["\n".join([str(i) for i in (a, b) if i])
                       for a, b in zip(times_list, text_list)]

    if stacked_plot == True:
        


        # Create table for stacked plot
        stacked_plot_table = calc_class_ratio(da)

        # Build colour list of hex vals for stacked plot
        hex_colour_list = []
        colour_def = lc_colours[measurement]

        # Custom error message to help if user puts incorrect measurement name
        for val in list(stacked_plot_table):
            try:
                r, g, b = colour_def[val][0:3]
            except KeyError:
                raise KeyError(
                    "class number not found in colour definition. "
                    "Ensure measurement name provided matches the dataset being used")
            hex_val = rgb_to_hex(r, g, b)
            hex_colour_list.append(hex_val)

        # Define & set up figure
        fig, (ax1, ax2) = plt.subplots(1, 2, dpi=dpi, constrained_layout=True)
        fig.set_size_inches(width * scale * 2, height * scale, forward=True)
        fig.set_constrained_layout_pads(
            w_pad=0.2, h_pad=0.2, hspace=0, wspace=0)

        # This function is called at regular intervals with changing i
        # values for each frame
        def _update_frames(i, ax1, ax2, extent, annotation_text,
                           annotation_defaults, cmap, norm):
            # Clear previous frame to optimise render speed and plot imagery
            ax1.clear()
            ax2.clear()

            ax1.imshow(da[i, ...], cmap=cmap, norm=norm,
                       extent=extent, interpolation="nearest")
            if(not label_ax):
                ax1.set_axis_off()

            clipped_table = stacked_plot_table.iloc[: int(i + 1)]
            data = clipped_table.to_dict(orient="list")
            date = clipped_table.index

            ax2.stackplot(date, data.values(), colors=hex_colour_list)
            ax2.tick_params(axis="x", labelrotation=-45)
            ax2.margins(x=0, y=0)

            # Add annotation text
            ax1.annotate(annotation_text[i], **annotation_defaults)
            ax2.annotate(annotation_text[i], **annotation_defaults)

        # anim_fargs contains all the values we send to our
        # _update_frames function.
        # Note the layer_cmap and layer_norm which were calculated
        # earlier being passed through
        anim_fargs = (
            ax1,
            ax2,  # axis to plot into
            [left, right, bottom, top],  # imshow extent
            annotation_list,
            annotation_defaults,
            layer_cmap,
            layer_norm,
        )

    else:  # stacked_plot = False

        # if plotting level 4 with colourbar

        if measurement == 'level4' and colour_bar == True:

            # specific setting to fit level 4 colour bar beside the plot
            # we will plot the animation in the left hand plot
            # and put the colour bar on the right hand side

            # Define & set up figure, two subplots so colour bar fits :)
            fig, (ax1, ax2) = plt.subplots(1, 2, dpi=dpi,
                                           constrained_layout=True, gridspec_kw={'width_ratios': [3, 1]})
            fig.set_size_inches(width * scale * 2,
                                height * scale, forward=True)
            fig.set_constrained_layout_pads(
                w_pad=0.2, h_pad=0.2, hspace=0, wspace=0)

            # make colour bar
            # provide left hand canvas to colour bar fuction which is where the image will go
            # colourbar will plot on right side beside it

            make_colorbar(fig, ax1, measurement, animation=True)

            # turn off lines for second plot so it's not ontop of colourbar
            ax2.set_axis_off()

        # plotting any other measurement with or with-out colour bar or level 4 without
        else:

            # Define & set up figure
            fig, ax1 = plt.subplots(1, 1, dpi=dpi)
            fig.set_size_inches(width * scale, height * scale, forward=True)
            if(not label_ax):
                fig.subplots_adjust(left=0, bottom=0, right=1,
                                    top=1, wspace=None, hspace=None)
            # Add colourbar here
            if colour_bar:
                make_colorbar(fig, ax1, measurement)


        # This function is called at regular intervals with changing i
        # values for each frame
        def _update_frames(i, ax1, extent, annotation_text,
                           annotation_defaults, cmap, norm):
            # Clear previous frame to optimise render speed and plot imagery
            ax1.clear()
            ax1.imshow(da[i, ...], cmap=cmap, norm=norm,
                       extent=extent, interpolation="nearest")
            if(not label_ax):
                ax1.set_axis_off()

            # Add annotation text
            ax1.annotate(annotation_text[i], **annotation_defaults)

        # anim_fargs contains all the values we send to our
        # _update_frames function.
        # Note the layer_cmap and layer_norm which were calculated
        # earlier being passed through
        anim_fargs = (
            ax1,
            [left, right, bottom, top],  # imshow extent
            annotation_list,
            annotation_defaults,
            layer_cmap,
            layer_norm,
        )

    # Animate
    anim = FuncAnimation(
        fig=fig,
        func=_update_frames,
        fargs=anim_fargs,
        frames=len(da.time),
        interval=animation_interval,
        repeat=False,
    )

    anim.save(file_name, writer="pillow", dpi=dpi)
    plt.close()
    return Image(filename=file_name)


"""
Description: module to allow tide height calculations with the FES2014 model
using the [CNES aviso-fes](https://github.com/CNES/aviso-fes) Python
implementation. It is coded to be used as a drop-in replacement for the default
OTPS-based estimations in coastal.tidal_tag.

This module requires the Aviso FES2014 calibration data to be obtained from
Aviso and extracted in a location and the corresponding (normal or extrapolated)
`.ini` file path to be exported in the `FES2014_OCEAN_INI` environment variable.

While not as flexible in terms of model choice as the `pyTMD` approach proposed
in
[DEAfrica](https://github.com/GeoscienceAustralia/dea-coastlines/blob/deafrica/deafrica_coastlines/raster.py),
the `aviso-fes` proved to be dramatically faster.
""" 

from os import environ
from pathlib import Path
from types import SimpleNamespace

import numpy as np

from pyfes import Handler


def initialise_handler():
    """Initialise Aviso FES Handler from INI file.

    The path to the INI file must be set in the `FES2014_OCEAN_INI` environment
    variable.
    """
    ini_path = Path(environ.get("FES2014_OCEAN_INI", "."))
    if not ini_path.exists() or not ini_path.is_file():
        raise ValueError("FES2014_OCEAN_INI environment variable must be set")
    short_tide = Handler("ocean", "io", str(ini_path))
    print(f"Initialised FES2014 from {ini_path}")
    return short_tide

def predict_tide(timepoints):
    """Compute tide heights using the pyfes handler.

    This method computes the tide heights at a given (latitude, longitude) for a
    set of naive UTC times array. These should be packaged as a list of
    `TimePoint` i.e., tuples of (lon, lat, time).

    It returns a list of SimpleNamespaces, each containing the "pure tide" as
    seen by a tide gauge, expressed in meters at the corresponding time.
    This is the sum of the computed height of the diurnal and semi-diurnal
    constituents of the tidal spectrum and of the long period wave constituents
    of the tidal spectrum.
    """        
    lons, lats, times = tuple(np.array(timepoints).T)
    # aviso-fes requires naive UTC times in microseconds
    times_us = times.astype("datetime64[us]")
    tide, lp, _ = SHORT_TIDE.calculate(lons, lats, times_us)

    # Heights correspond to tide + lp converted from cm to m
    heights = (tide + lp) / 100
    # Package result to be compatible with OTPS, as needed by coastal.py
    return [SimpleNamespace(tide_m=val) for val in heights]


def TimePoint(*args):
    """Dummy method to package data as requied by `coastal`."""
    return args


# Initialise Aviso FES handler
SHORT_TIDE = initialise_handler()
